﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using SimpleJSON;
using UnityEngine;
using UnityEngine.Events;
using UnityEngine.SceneManagement;
using UnityEngine.Serialization;
using UnityEngine.UI;
using Object = UnityEngine.Object;

namespace AudioMate {
    /// <summary>
    /// AudioMate by dub
    /// -------------------------------------------------------
    /// Virt-a-Mate plugin for audio clip playback
    /// and audio clip organisation
    /// -------------------------------------------------------
    /// https://github.com/dub-a-licious/audio-mate
    /// </summary>
    public class AudioMateController : MVRScript
    {
        public static string GuessInitialReceivingNode(Atom atom)
        {
            switch (atom.category)
            {
                case "People":
                    return "HeadAudioSource";
                case "Sound":
                    switch (atom.type)
                    {
                        case "AudioSource":
                            return "AudioSource";
                        case "RhythmAudioSource":
                            return "RhythmSource";
                        case "AptSpeaker":
                            return "AptSpeaker_Import";
                        default:
                            return "None";
                    }
                default:
                    return "None";
            }
        }
        public class ContainingAtomNameChangedEventArgs { public string Before; public string After; }
        public class ContainingAtomNameChangedEvent : UnityEvent<ContainingAtomNameChangedEventArgs> {};
        public readonly ContainingAtomNameChangedEvent OnContainingAtomNameChanged = new ContainingAtomNameChangedEvent();
        public class PluginInitializedEvent : UnityEvent {};
        public readonly PluginInitializedEvent OnPluginInitialized = new PluginInitializedEvent();
        public class BulkModeOffEvent : UnityEvent {};
        public readonly BulkModeOffEvent OnBulkModeOff = new BulkModeOffEvent();


        private Scene _activeScene;

        public CollectionManager collections;
        [FormerlySerializedAs("audioMateClipLibrary")] public ClipLibrary clipLibrary;
        public UIManager ui;
        public new MVRPluginManager Manager => base.manager;
        [FormerlySerializedAs("fileManager")] public AudioFileManager audioFileManager;

        public Atom receivingAtom;
        public JSONStorable receivingNode;

        public bool needsProvisionalClipProcessing;
        public bool uiInitialized;
        private bool _isBulkSelectOn;

        public bool isBulkSelectOn
        {
            get { return _isBulkSelectOn; }
            set
            {
                if (_isBulkSelectOn == value) return;
                _isBulkSelectOn = value;
                if (_isBulkSelectOn == false) OnBulkModeOff.Invoke();
            }
        }
        public int bulkSelectAmount;
        private bool _initializingUI;
        private bool _pluginInitialized;

        private string _missingReceiverNodeStoreId = "";
        private string _atomCreationId;

        // Active Collection Actions
        private JSONStorableAction _playClipActionJSON;
        private JSONStorableAction _queueClipActionJSON;
        private JSONStorableAction _playRandomClipActionJSON;
        private JSONStorableAction _queueRandomClipActionJSON;
        private JSONStorableAction _playNextClipActionJSON;
        private JSONStorableAction _playPreviousClipActionJSON;
        private JSONStorableAction _playCurrentClipActionJSON;
        private JSONStorableAction _setPlayIndexActionJSON;
        private JSONStorableAction _resetPlayIndexActionJSON;
        private JSONStorableAction _enableActionJSON;
        private JSONStorableAction _disableActionJSON;

        // Storables
        public JSONStorableStringChooser CollectionsJSON;
        public JSONStorableStringChooser ReceivingAtomJSON;
        public JSONStorableStringChooser ReceivingNodeJSON;
        public JSONStorableStringChooser TriggerColliderChooserJSON;
        public JSONStorableStringChooser MorphChooserJSON;
        public JSONStorableFloat FirstMultiselectButtonAmountJSON;
        public JSONStorableFloat SecondMultiselectButtonAmountJSON;

        // Global Settings
        public JSONStorableBool DisplayHintsJSON;
        public JSONStorableBool VolumeControlEnabledJSON;
        public JSONStorableBool PitchControlEnabledJSON;
        public JSONStorableBool AdvancedUIEnabledJSON;
        public JSONStorableBool AddImportedFilesToActiveCollectionJSON;

        public ClipCursor Cursor => clipLibrary.Cursor;
        public List<AudioMateClip> Clips => clipLibrary.Clips;
        public bool DisplayHints => DisplayHintsJSON.val;

        private Dictionary<string, Trigger> _collisionTriggerCache;

        #region Initialization
        public override void Init()
        {
            Log("Initializing Main Controller");
            try
            {
                _activeScene = SceneManager.GetActiveScene();
                base.Init();
                InitStorables();
                InitFileManager();
                StartCoroutine(DeferredInit());
                InitCollisionTriggerCache();
            }
            catch (Exception e)
            {
                SuperController.LogError($"AudioMate.{nameof(AudioMateController)}.{nameof(Init)}: {e}");
            }
        }

        protected override void Awake()
        {
            SuperController.singleton.onSceneLoadedHandlers += OnSceneLoaded;
            base.Awake();
        }

        private void OnSceneLoaded()
        {
            // if (this != null && collections != null) containingAtom.RestoreFromLast(this);
        }

        private IEnumerator DeferredInit()
        {
            Log("Starting deferred initialization routine");
            if ((Object) collections == (Object) null)
            {
                while (_activeScene.isLoaded == false || SuperController.singleton.isLoading)
                {
                    Log("Waiting for scene to finish loading...");
                    yield return null;
                    if ((Object) this == (Object) null) yield break;
                }

                while (!uiInitialized)
                {
                    Log("Waiting for UI to be initialized...");
                    yield return new WaitForSeconds(.1f);
                    if ((Object) this == (Object) null) yield break;
                }
                InitCollections();
                InitClipLibrary();
                clipLibrary.Bind();
            }

            Log("Trying to restore from saved scene data.");
            containingAtom.RestoreFromLast(this);
        }

        public override void InitUI()
        {
            Log("Initializing UI");

            base.InitUI();
            try
            {
                if (UITransform == null) return;

                SuperController.singleton.StartCoroutine(InitUIDeferred());
            }
            catch (Exception exc)
            {
                SuperController.LogError($"AudioMate.{nameof(AudioMateController)}.{nameof(InitUI)}: {exc}");
            }
        }

        private IEnumerator InitUIDeferred()
        {
            if ((Object) ui != (Object) null || _initializingUI) yield break;
            _initializingUI = true;
            yield return SuperController.singleton.StartCoroutine(UIFactory.LoadUIAssets());

            var scriptUI = UITransform.GetComponentInChildren<MVRScriptUI>();
            var scrollRect = scriptUI.fullWidthUIContent.transform.parent.parent.parent.GetComponent<ScrollRect>();

            if (scrollRect == null)
                SuperController.LogError(
                    "AudioMate: Error during UI initialization. Scroll rect not at the expected hierarchy position");
            else
            {
                scrollRect.vertical = false;
                scrollRect.horizontal = false;
                scrollRect.elasticity = 0;
                scrollRect.inertia = false;
                scrollRect.movementType = ScrollRect.MovementType.Clamped;
            }

            ui = UIManager.AddTo(scriptUI.fullWidthUIContent);
            ui.OnUIInitialized.AddListener(OnUIInitialized);
            ui.Init(this);
            var count = 0;
            while ((Object) collections == (Object) null || !collections.isInitialized)
            {
                // security break out to avoid endless loops
                if (++count > 1000)
                {
                    SuperController.LogError("Security break out in InitUIDeferred.");

                    yield break;
                }
                yield return new WaitForSeconds(.1f);
                if ((Object) this == (Object) null) yield break;
            }
            try
            {
                ui.Bind();
            }
            catch (Exception e)
            {
                SuperController.LogError($"InitUIDeferred: {e}");
            }
        }

        public void OnUIInitialized()
        {
            Log("Initialized UI");
            uiInitialized = true;
        }

        private void InitCollisionTriggerCache()
        {
            if (_collisionTriggerCache == null)
            {
                _collisionTriggerCache = new Dictionary<string, Trigger>();
            }
            if (_collisionTriggerCache?.Count > 0)
            {
                _collisionTriggerCache.Clear();
            }
            foreach (var triggerUID in ColliderChoices)
            {
                var collisionTrigger = containingAtom.GetStorableByID(triggerUID) as CollisionTrigger;
                if (collisionTrigger == null) return;
                _collisionTriggerCache?.Add(triggerUID, collisionTrigger.trigger);
            }
        }

        private void InitCollections()
        {
            if ((UnityEngine.Object) collections != (UnityEngine.Object) null) return;
            Log("Initializing Collections");
            try
            {
                collections = gameObject.AddComponent<CollectionManager>();
                collections.transform.SetParent(transform, false);

                collections.OnActiveCollectionSelected.RemoveListener(OnActiveCollectionSelected);
                collections.OnActiveCollectionSelected.AddListener(OnActiveCollectionSelected);

                // Active Collection Actions
                _playClipActionJSON = new JSONStorableAction(Storables.PlayClipAction,
                    () => collections.PlayClipAction());
                RegisterAction(_playClipActionJSON);

                _queueClipActionJSON = new JSONStorableAction(Storables.QueueClipAction,
                    () => collections.QueueClipAction());
                RegisterAction(_queueClipActionJSON);

                _playRandomClipActionJSON = new JSONStorableAction(Storables.PlayRandomClipAction,
                    () => collections.PlayRandomClipAction());
                RegisterAction(_playRandomClipActionJSON);

                _queueRandomClipActionJSON = new JSONStorableAction(Storables.QueueRandomClipAction,
                    () => collections.QueueRandomClipAction());
                RegisterAction(_queueRandomClipActionJSON);

                _playNextClipActionJSON = new JSONStorableAction(Storables.PlayNextClipAction,
                    () => collections.PlayNextClipAction());
                RegisterAction(_playNextClipActionJSON);

                _playPreviousClipActionJSON = new JSONStorableAction(Storables.PlayPreviousClipAction,
                    () => collections.PlayPreviousClipAction());
                RegisterAction(_playPreviousClipActionJSON);

                _playCurrentClipActionJSON = new JSONStorableAction(Storables.PlayCurrentClipAction,
                    () => collections.PlayCurrentClipAction());
                RegisterAction(_playCurrentClipActionJSON);

                _resetPlayIndexActionJSON = new JSONStorableAction(Storables.ResetPlayIndexAction,
                    () => collections.ResetPlayIndexAction());
                RegisterAction(_resetPlayIndexActionJSON);

                _enableActionJSON = new JSONStorableAction(Storables.EnableActiveCollectionAction,
                    () => collections.Enable());
                RegisterAction(_enableActionJSON);

                _disableActionJSON = new JSONStorableAction(Storables.DisableActiveCollectionAction,
                    () => collections.Disable());
                RegisterAction(_disableActionJSON);

                collections.Init(this);
            }
            catch (Exception e)
            {
                SuperController.LogError($"AudioMate.{nameof(AudioMateController)}.{nameof(InitCollections)}: {e}");
            }
        }

        private void InitFileManager()
        {
            Log("Initializing File Manager");
            audioFileManager = gameObject.AddComponent<AudioFileManager>();
        }

        private void InitClipLibrary()
        {
            Log("Initializing Clip Library");
            if ((Object) clipLibrary == (Object) null) clipLibrary = gameObject.AddComponent<ClipLibrary>();
            clipLibrary.Init(this);
        }

        private void InitStorables()
        {
            Log("Initializing Storables");

            CollectionsJSON = new JSONStorableStringChooser(
                Storables.ActiveCollection,
                new List<string>(),
                null,
                "Active Collection")
            {
                storeType = JSONStorableParam.StoreType.Full
            };
            RegisterStringChooser(CollectionsJSON);

            ReceivingAtomJSON = new JSONStorableStringChooser(Storables.ReceivingAtom, null,
                containingAtom.uid, "Receiver Atom", SetReceivingAtom)
            {
                isRestorable = false,
                isStorable = false
            };
            RegisterStringChooser(ReceivingAtomJSON);

            ReceivingNodeJSON = new JSONStorableStringChooser(Storables.ReceivingNode, null,
                GuessInitialReceivingNode(containingAtom), "Receiver Node", SetReceivingNode)
            {
                isRestorable = false,
                isStorable = false
            };
            RegisterStringChooser(ReceivingNodeJSON);

            TriggerColliderChooserJSON = new JSONStorableStringChooser(
                Storables.TriggerActions,
                ColliderChoices,
                DefaultCollider,
                "Trigger")
            {
                isRestorable = false,
                isStorable = false
            };
            RegisterStringChooser(TriggerColliderChooserJSON);

            MorphChooserJSON = new JSONStorableStringChooser(
                    Storables.Morphs,
                    GetMorphLabelList(),
                    "None",
                    "Morph")
            {
                isRestorable = false,
                isStorable = false
            };
            // RegisterStringChooser(MorphChooserJSON);

            // Global Settings
            AdvancedUIEnabledJSON = new JSONStorableBool(
                Storables.AdvancedUIEnabled,
                false);
            RegisterBool(AdvancedUIEnabledJSON);

            DisplayHintsJSON = new JSONStorableBool(
                Storables.DisplayHints,
                false);
            RegisterBool(DisplayHintsJSON);

            VolumeControlEnabledJSON = new JSONStorableBool(
                Storables.VolumeControlEnabled,
                false,
                val => collections.SetVolumeControl(val));
            RegisterBool(VolumeControlEnabledJSON);

            PitchControlEnabledJSON = new JSONStorableBool(
                Storables.PitchControlEnabled,
                false,
                val => collections.SetPitchControl(val));
            RegisterBool(PitchControlEnabledJSON);

            AddImportedFilesToActiveCollectionJSON = new JSONStorableBool(
                Storables.AddImportedFilesToActiveCollectionEnabled,
                false);
            RegisterBool(AddImportedFilesToActiveCollectionJSON);

            FirstMultiselectButtonAmountJSON = new JSONStorableFloat(
                Storables.FirstMultiselectButtonAmount, 10, 2, 100);
            RegisterFloat(FirstMultiselectButtonAmountJSON);

            SecondMultiselectButtonAmountJSON = new JSONStorableFloat(
                Storables.SecondMultiselectButtonAmount, 20, 2, 100);
            RegisterFloat(SecondMultiselectButtonAmountJSON);

            // Bind onAtomUIDRename event
            SuperController.singleton.onAtomUIDRenameHandlers += OnAtomRename;
        }

        private readonly List<string> _colliderChoices = new List<string>();
        public List<string> ColliderChoices
        {
            get {
                _colliderChoices.Clear();
                _colliderChoices.Add("LipTrigger");
                _colliderChoices.Add("MouthTrigger");
                _colliderChoices.Add("ThroatTrigger");
                _colliderChoices.Add("lNippleTrigger");
                _colliderChoices.Add("rNippleTrigger");
                _colliderChoices.Add("LabiaTrigger");
                _colliderChoices.Add("VaginaTrigger");
                _colliderChoices.Add("DeepVaginaTrigger");
                _colliderChoices.Add("DeeperVaginaTrigger");
                // Add custom collision triggers
                _colliderChoices.AddRange(from atom in SuperController.singleton.GetAtoms()
                    where atom != null
                    where atom.type == "CollisionTrigger"
                    select atom.name);
                return _colliderChoices;
            }
        }

        public const string DefaultCollider = "VaginaTrigger";

        #endregion

        #region Lifecycle
        public void OnEnable()
        {
            if (collections == null) return;
            collections.EnableAll();
        }

        public void OnDisable()
        {
            if (collections != null) collections.DisableAll();
        }

        public void Start()
        {
            Log("Start");
        }

        public void OnDestroy()
        {
            Log("Destroying Plugin instance");
            SuperController.singleton.onSceneLoadedHandlers -= OnSceneLoaded;
            SuperController.singleton.onAtomUIDRenameHandlers -= OnAtomRename;
            OnContainingAtomNameChanged.RemoveAllListeners();
            OnPluginInitialized.RemoveAllListeners();
            if (ui != null)
            {
                ui.OnDestroy();
                Destroy(ui);
            }
            if (collections == null) return;
            collections.OnDestroy();
            Destroy(collections);
        }

        public void Update()
        {
            if (((Object) SuperController.singleton != (Object) null && SuperController.singleton.freezeAnimation)
                || enabled == false
                || !uiInitialized
                || (Object) collections == (Object) null) return;
            if (_pluginInitialized)
            {
                collections.Update();
            }
            if ((Object) clipLibrary != (Object) null && needsProvisionalClipProcessing && clipLibrary.isInitialised)
            {
                needsProvisionalClipProcessing = false;
                clipLibrary.RefreshClipIndex();
            }
            FindMissingReceiverNode();
            if (_pluginInitialized
                || !uiInitialized
                || (Object) collections == (Object) null
                || collections.ActiveCollection == null
                || (Object) clipLibrary == (Object) null
                || !clipLibrary.isInitialised
                || collections.isParsingCollections) return;
            _pluginInitialized = true;
            Log("### Plugin initialized ###");
            OnPluginInitialized.Invoke();
            collections.StartAutoPlayRoutine();
        }

        public void FixedUpdate()
        {
            if ((Object) ui != (Object) null)
            {
                ui.FixedUpdate();
            }

            if ((Object) SuperController.singleton != (Object) null &&
                SuperController.singleton.freezeAnimation) return;
            if (_collisionTriggerCache == null) return;
            foreach (var keyValuePair in _collisionTriggerCache)
            {
                foreach (var collection in collections)
                {
                    if (collection.Enabled) collection.SetCollisionTriggerActiveState(keyValuePair.Key, keyValuePair.Value.active);
                }
            }
        }

        /**
		 * Create/refresh the list of available custom scene clips.
		 */
        public void RefreshClipLibrary()
        {
            if ((UnityEngine.Object) clipLibrary == (UnityEngine.Object) null) return;
            clipLibrary.RefreshClipIndex();
        }

        private void OnAtomRename(string oldID, string newID)
        {
            if (newID != containingAtom.uid) return;
            Log($"Containing Atom has been renamed to {newID}");
            if ((UnityEngine.Object) collections == (UnityEngine.Object) null) return;
            OnContainingAtomNameChanged.Invoke(new ContainingAtomNameChangedEventArgs
            {
                Before = oldID,
                After = newID
            });
        }
        #endregion

        #region Load / Save
        public override JSONClass GetJSON(bool includePhysical = true, bool includeAppearance = true, bool forceStore = false)
        {
            var json = base.GetJSON(includePhysical, includeAppearance, forceStore);

            if (collections == null) return json;
            collections.AddToJSON(json, Storables.Collections);
            needsStore = true;

            return json;
        }
        public override void RestoreFromJSON(JSONClass jc, bool restorePhysical = true, bool restoreAppearance = true, JSONArray presetAtoms = null, bool setMissingToDefault = true)
        {
            if (jc == null || !jc.Childs.Any()) return;
            base.RestoreFromJSON(jc, restorePhysical, restoreAppearance, presetAtoms, setMissingToDefault);
            try
            {
                if (!jc.HasKey(Storables.Collections))
                {
                    Log("No processable data found. Aborting restore.");
                    return;
                }
                var activeCollection = jc.HasKey(Storables.ActiveCollection)
                    ? jc[Storables.ActiveCollection].Value
                    : null;

                StartCoroutine(LoadDeferred(jc, activeCollection));
            }
            catch (Exception e)
            {
                SuperController.LogError($"AudioMate.{nameof(AudioMateController)}.{nameof(RestoreFromJSON)}: {e}");
            }
        }

        public override void LateRestoreFromJSON(JSONClass jc, bool restorePhysical = true, bool restoreAppearance = true,
            bool setMissingToDefault = true)
        {
            // if (hasRestoredFromJSON) return;
            Log($"Late Restore");
            if (!jc.AsObject.HasKey(Storables.Collections))
            {
                Log("No processable data found. Aborting late restore.");
                return;
            }
            base.LateRestoreFromJSON(jc, restorePhysical, restoreAppearance, setMissingToDefault);
            RestoreFromJSON(jc, restorePhysical, restoreAppearance, null, setMissingToDefault);
        }

        private IEnumerator LoadDeferred(JSONNode json, string activeCollection = null)
        {
            if (insideRestore) yield break;
            insideRestore = true;
            if (json == null || json.AsObject == null || !json.Childs.Any())
            {
                Log($"Load called with invalid json object: {json}. Aborting load!");
                insideRestore = false;
                yield break;
            }
            while ((Object) collections == (Object) null)
            {
                // Wait for collections to be initialized
                yield return null;
            }
            try
            {
                Log($"Parsing collections.");
                if (json.AsObject?.HasKey(Storables.Collections) != true) yield break;
                var jsonCollections = json[Storables.Collections];
                if (jsonCollections == null) yield break;
                Log("Found collection data. Parsing into collections...");
                collections.RestoreFromJSON(jsonCollections, false, false, activeCollection);
            }
            catch (Exception e)
            {
                SuperController.LogError($"AudioMate.{nameof(AudioMateController)}.{nameof(LoadDeferred)}: {e}");
            }
            finally
            {
                insideRestore = false;
            }
        }
        #endregion

        #region Helpers

        public bool IsInterferingAutoBehaviourEnabled()
        {
            return containingAtom.GetStorableByID("JawControl").GetBoolParamValue("driveXRotationFromAudioSource") ||
                   containingAtom.GetStorableByID("AutoJawMouthMorph").GetBoolParamValue("enabled");
        }

        public void DisableInterferingPersonAtomSettings()
        {
            containingAtom.GetStorableByID("JawControl").SetBoolParamValue("driveXRotationFromAudioSource", false);
            containingAtom.GetStorableByID("AutoJawMouthMorph").SetBoolParamValue("enabled", false);
        }

        public List<string> GetMorphLabelList()
        {
            if (containingAtom == null) return new List<string>();
            var storable = containingAtom.GetStorableByID("geometry");
            if (storable == null) return new List<string>();
            return storable.GetFloatParamNames() ?? new List<string>();
        }

        public JSONStorableFloat GetMorph(string morphName)
        {
            if ((Object) containingAtom == (Object) null) return null;
            var storable = containingAtom.GetStorableByID("geometry");
            return (Object) storable == (Object) null ? null : storable.GetFloatJSONParam(morphName);
        }

        public virtual IEnumerator CreateAtom(string atomType, string atomId, Action<Atom> onAtomCreated)
        {
            if (_atomCreationId != null && _atomCreationId == atomId) yield break;
            _atomCreationId = atomId;
            var atom = SuperController.singleton.GetAtomByUid(atomId);
            if ((Object) atom == (Object) null)
            {
                yield return SuperController.singleton.AddAtomByType(atomType, atomId);
                atom = SuperController.singleton.GetAtomByUid(atomId);
            }
            if ((Object) atom != (Object) null)
            {
                onAtomCreated(atom);
            }
            yield return null;
            _atomCreationId = null;
        }

        public void ParentLinkAtPosition(Atom atom, Atom parent, JSONStorable toPosition, string toPositionName)
        {
            atom.parentAtom = parent;
            var p = toPosition.transform.position;
            atom.mainController.currentPositionState = FreeControllerV3.PositionState.On;
            atom.mainController.transform.position = new Vector3(p.x, p.y, p.z);
            atom.mainController.SetLinkToAtom(parent.uid);
            atom.mainController.SetLinkToRigidbodyObject(toPositionName);
            atom.mainController.currentPositionState = FreeControllerV3.PositionState.ParentLink;
            atom.mainController.canGrabPosition = false;
            atom.mainController.canGrabRotation = false;
        }

        #endregion

        #region Callbacks
        public void OnActiveCollectionSelected(CollectionManager.ActiveCollectionSelectedEventArgs eventArgs)
        {
            if ((Object) collections == (Object) null) return;

            var atom = collections.ActiveCollection?.ReceiverAtomUID;
            if (atom == null) return;
            ReceivingAtomJSON.val = atom;

            var node = collections.ActiveCollection?.ReceiverNodeUID;
            if (node == null) return;
            ReceivingNodeJSON.val = node;
        }

        public void OnActiveCollectionNameChanged(CollectionManager.ActiveCollectionNameChangedEventArgs eventArgs)
        {
            // triggerManager.UpdateTriggerActionNamesOnCollectionRename(eventArgs.Before, eventArgs.After);
            // TODO is there anything to do here?
        }

        /**
		* Load audio source ReceiverAtom.
		*/
        private void SetReceivingAtom(string uid)
        {
            if (Tools.IsEmptyChoice(uid) || (UnityEngine.Object) collections == (UnityEngine.Object) null || collections.ActiveCollection == null) return;
            try
            {
                receivingAtom = SuperController.singleton.GetAtomByUid(uid);
                if (receivingAtom == null)
                {
                    Log($"Audio Source {uid} does not exist.");
                    return;
                }
                var atomChanged = collections.ActiveCollection.ReceiverAtomUID != receivingAtom.uid;
                collections.ActiveCollection.ReceiverAtomUID = receivingAtom.uid;
                if (!atomChanged) return;
                var guessedReceiver = GuessInitialReceivingNode(receivingAtom);
                insideRestore = true;
                ReceivingNodeJSON.val = guessedReceiver;
                insideRestore = false;
            }
            catch (Exception e)
            {
                SuperController.LogError($"AudioMate.{nameof(AudioMateController)}.{nameof(SetReceivingAtom)}: {e}");
            }
        }

        /**
		* Load audio source receiver.
		*/
        private void SetReceivingNode(string uid)
        {
            if (Tools.IsEmptyChoice(uid) || (UnityEngine.Object) collections == (UnityEngine.Object) null || collections.ActiveCollection == null) return;
            receivingNode = SuperController.singleton.GetAtomByUid(ReceivingAtomJSON.val)?.GetStorableByID(uid);
            if ((Object) receivingNode == (Object) null)
            {
                _missingReceiverNodeStoreId = uid;
                return;
            }

            insideRestore = true;
            ReceivingNodeJSON.val = receivingNode.name;
            insideRestore = false;
            collections.ActiveCollection.ReceiverNodeUID = receivingNode.name;
            collections.ActiveCollection.SyncAudioReceiver();
        }

        private void FindMissingReceiverNode()
        {
            if (_missingReceiverNodeStoreId == "" || (Object) receivingAtom == (Object) null) return;
            var missingReceiver = receivingAtom.GetStorableByID(_missingReceiverNodeStoreId);
            if ((Object) missingReceiver == (Object) null) return;
            SetReceivingNode(_missingReceiverNodeStoreId);
            _missingReceiverNodeStoreId = "";
        }
        #endregion

        #region Debug
        /**
         * Custom debug logging
         */
        public void Log(string message)
        {
            #if DEBUG
            SuperController.LogMessage($"[{this.storeId}]: {message}");
            #endif
        }

        /* public void LogObject(object obj)
        {
            #if DEBUG
            SuperController.LogMessage(JsonConvert.SerializeObject(obj));
            #endif
        } */

        /**
         * UI debugging. Depends on VAM Dev Tools by Acidbubbles (https://github.com/acidbubbles/vam-devtools).
         */
        public void DebugObject(Transform obj)
        {
            #if DEBUG
            DebugObject(obj.gameObject);
            #endif
        }

        public void DebugObject(GameObject obj)
        {
            #if DEBUG
            containingAtom.BroadcastMessage("DevToolsInspectUI", obj);
            #endif
        }

        public void DebugUI(GameObject go)
        {
            var components = go.GetComponentsInChildren<Component>();
            var componentsCount = components.Length;
            var rectTransform = go.GetComponent<RectTransform>();
            Log($"DEBUGGING UI ELEMENT '{go.name}'");
            foreach (var component in components)
            {
                Log($"{component.name}: {component.GetType()}");
                if (component.transform.parent != null) {
                    Log($"Parent: {component.transform.parent.gameObject.name}");
                }
            }
        }
        #endregion
    }
}
