﻿using System;
using System.ComponentModel;
using System.Linq;
using MVR.FileManagement;
using SimpleJSON;
using UnityEngine;
using UnityEngine.Events;
using UnityEngine.EventSystems;
using UnityEngine.UI;
using Object = UnityEngine.Object;

namespace AudioMate
{
    public class ClipUI
    {
        public bool IsInitialised;
        public GameObject GameObject;
        public Transform Container;
        public UIDynamicButton PreviewButton;
        public UIDynamicButton ToggleButton;
        public UIDynamicButton RemoveButton;
        public Outline ToggleOutline;
        public MouseOverDetector MouseOver;

        public bool SetInCollectionState(bool state)
        {
            if ((Object) ToggleButton == (Object) null) return false;
            ToggleButton.textColor = state ? Styles.EnabledText : Styles.DisabledText;
            ToggleButton.buttonColor = state ? Styles.EnabledBg : Styles.DisabledBg;
            if ((Object) ToggleOutline != (Object) null) ToggleOutline.effectColor = state ? Color.white : Styles.DisabledBg;

            return state;
        }

        public bool PreviewInCollectionState(bool state)
        {
            if ((Object) ToggleButton == (Object) null) return false;
            ToggleButton.textColor = state ? Styles.EnabledText : Styles.DisabledText;
            ToggleButton.buttonColor = state ? Styles.EnabledBg : Styles.DisabledBg;
            if ((Object) ToggleOutline != (Object) null) ToggleOutline.effectColor = state ? Color.white : Styles.DisabledBg;

            return state;
        }

        public bool SetCursorState(bool state)
        {
            if ((Object) PreviewButton == (Object) null) return state;
            PreviewButton.textColor = state ? Styles.CursorEnabledText : Styles.ButtonText;
            PreviewButton.buttonColor = state ? Styles.CursorEnabledBg : Styles.ButtonBg;
            return state;
        }

        public void Destroy()
        {
            try
            {
                if ((Object) Container != (Object) null) Object.Destroy(Container.gameObject);
                Container = null;
                if ((Object) PreviewButton != (Object) null) Object.Destroy(PreviewButton.gameObject);
                PreviewButton = null;
                if ((Object) ToggleButton != (Object) null) Object.Destroy(ToggleButton.gameObject);
                ToggleButton = null;
                if ((Object) ToggleOutline != (Object) null) Object.Destroy(ToggleOutline.gameObject);
                ToggleOutline = null;
                if ((Object) RemoveButton != (Object) null) Object.Destroy(RemoveButton.gameObject);
                RemoveButton = null;
                if ((Object) MouseOver != (Object) null)
                {
                    MouseOver.OnMouseEntered.RemoveAllListeners();
                    MouseOver.OnMouseExited.RemoveAllListeners();
                    Object.Destroy(MouseOver.gameObject);
                }
                MouseOver = null;
                IsInitialised = false;
            }
            catch (Exception e)
            {
                SuperController.LogError($"AudioMate.{nameof(ClipUI)}.{nameof(Destroy)}: {e}");
            }
        }
    }

    public class AudioMateClip
    {
        public string DisplayName => AudioClip?.displayName ?? ProvisionalDisplayName;
        public string ProvisionalDisplayName;
        public class ClipMouseOverStateChangedEventArgs { public AudioMateClip Clip; public bool IsOver; }
        public class ClipMouseOverStateChangedEvent : UnityEvent<ClipMouseOverStateChangedEventArgs> {};
        public readonly ClipMouseOverStateChangedEvent OnMouseEntered = new ClipMouseOverStateChangedEvent();
        public readonly ClipMouseOverStateChangedEvent OnMouseExited = new ClipMouseOverStateChangedEvent();
        public string UID
        {
            get {
                if (_uid != null || AudioClip == null) return _uid;
                _uid = $"AudioMateClip_{Tools.TrimFileExtension(AudioClip.uid)}";

                return _uid;
            }
        }

        private string _uid;
        public NamedAudioClip AudioClip { get; private set; }
        public AudioSourceControl Receiver;

        public bool IsInActiveCollection { get; private set; }
        public bool HasCursor { get; private set; }

        public ClipUI UI;

        public bool IsProvisional;

        public string SourcePath;

        public int CurrentIndex;

        public bool IsVisible => UI.GameObject.activeSelf;

        public AudioMateClip(NamedAudioClip audioClip)
        {
            AudioClip = audioClip;
            Init();

        }

        public AudioMateClip(JSONNode jn)
        {
            RestoreFromJSON(jn);
            Init();
        }

        private void Init()
        {
            UI = new ClipUI();
            IsInActiveCollection = false;
            HasCursor = false;
            AddSourcePath();
        }

        public ClipUI CreateClipUI(Transform parent, UnityAction previewCallback, UnityAction toggleCallback, UnityAction removeCallback)
        {
            var go = new GameObject(UID);
            go.transform.SetParent(parent, false);
            var rectTransform = go.AddComponent<RectTransform>();
            rectTransform.StretchParent();
            var layout = go.AddComponent<HorizontalLayoutGroup>();
            layout.childForceExpandWidth = false;
            layout.childControlWidth = true;
            layout.childForceExpandHeight = false;
            layout.childControlHeight = true;
            layout.padding = new RectOffset(0, 10, 0, 0);
            var buttonGroup = go.AddComponent<HorizontalButtonGroup>();

            UI.GameObject = go;
            UI.PreviewButton = buttonGroup.CreateButton(
                Symbols.Play,
                Styles.Button,
                previewCallback);
            UI.ToggleButton = buttonGroup.CreateButton(
                DisplayName,
                Styles.Disabled,
                toggleCallback,
                true,
                0,
                100f);
            UI.RemoveButton = buttonGroup.CreateButton(
                Symbols.Remove,
                Styles.Danger,
                removeCallback,
                true,
                50f,
                0);
            ;
            UI.MouseOver = UI.ToggleButton.gameObject.AddComponent<MouseOverDetector>();
            UI.MouseOver.OnMouseEntered.AddListener(() =>
            {
                OnMouseEntered.Invoke(new ClipMouseOverStateChangedEventArgs { Clip = this, IsOver = true });
            });
            UI.MouseOver.OnMouseExited.AddListener(() =>
            {
                OnMouseExited.Invoke(new ClipMouseOverStateChangedEventArgs { Clip = this, IsOver = false });
            });
            UI.IsInitialised = true;
            RefreshUI();
            return UI;
        }

        private void RefreshUI()
        {
            if (UI == null || UI.IsInitialised == false) return;
            UI.SetInCollectionState(IsInActiveCollection);
            UI.SetCursorState(HasCursor);
        }

        public JSONClass GetJSON()
        {
            return new JSONClass
            {
                { "uid", UID },
                { "audioClip", AudioClip?.uid },
                { "hasCursor", HasCursor.ToString() },
                { "sourcePath", SourcePath }
            };
        }

        private void RestoreFromJSON(JSONNode jn)
        {
            if (jn == null || !jn.Childs.Any()) return;
            var sourceClip = jn["audioClip"] ?? jn["sourceClip"] ?? jn["sourceClipUID"]; // Check for legacy formats
            SourcePath = jn["sourcePath"];
            AudioClip = sourceClip != null ? URLAudioClipManager.singleton.GetClip(sourceClip) : null;
            if (AudioClip == null && SourcePath != null)
            {
                AudioClip = AudioFileManager.Load(SourcePath);
            }
            if (string.IsNullOrEmpty(SourcePath))
            {
                AddSourcePath();
            }
        }

        private void AddSourcePath()
        {
            if (AudioClip == null) return;
            var sceneClipData = URLAudioClipManager.singleton.GetJSON()["clips"];
            foreach (var clipData in sceneClipData.Childs)
            {
                if (clipData["displayName"].Value == AudioClip.displayName)
                {
                    SourcePath = clipData["url"];
                }
            }
        }

        /**
         * Play the assigned audio clip with optional "if clear" param.
         */
        public bool Play(AudioSourceControl receiver = null, bool ifClear = false, bool clearQueue = false)
        {
            if ((Object) SuperController.singleton != (Object) null &&
                SuperController.singleton.freezeAnimation) return false;
            if (AudioClip == null) return false;
            if ((UnityEngine.Object) receiver != (UnityEngine.Object) null) Receiver = receiver;
            if (ifClear && Receiver.audioSource.isPlaying) return false;
            if (clearQueue)
            {
                Receiver.PlayNowClearQueue(AudioClip);
                return true;
            }

            Receiver.PlayNow(AudioClip);
            return true;
        }

        /**
         * Add assigned audio clip to audio _c queue list.
         */
        public void Queue(AudioSourceControl receiver = null)
        {
            if ((Object) SuperController.singleton != (Object) null &&
                SuperController.singleton.freezeAnimation) return;
            if (AudioClip == null) return;
            if ((UnityEngine.Object) receiver != (UnityEngine.Object) null)
            {
                Receiver = receiver;
            }
            else
            {
                return;
            }
            Receiver.QueueClip(AudioClip);
        }

        public void Hide()
        {
            UI?.GameObject.SetActive(false);
        }

        public void Show()
        {
            UI?.GameObject.SetActive(true);
        }

        public void DeleteFromScene()
        {
            if (AudioClip == null) return;
            URLAudioClipManager.singleton.RemoveClip(AudioClip);
            Hide();
            Destroy();
        }

        public void Destroy()
        {
            try
            {
                OnMouseEntered.RemoveAllListeners();
                OnMouseExited.RemoveAllListeners();
                UI.Destroy();
                Receiver = null;
                AudioClip = null;
            } catch (Exception e)
            {
                SuperController.LogError($"AudioMate.{nameof(AudioMateClip)}.{nameof(Destroy)}: {e}");
            }
        }

        public void DestroyUI()
        {
            try
            {
                UI.Destroy();
            } catch (Exception e)
            {
                SuperController.LogError($"AudioMate.{nameof(AudioMateClip)}.{nameof(DestroyUI)}: {e}");
            }
        }

        public bool ToggleState()
        {
            IsInActiveCollection = !IsInActiveCollection;
            RefreshUI();
            return IsInActiveCollection;
        }

        public void PreviewToggleState()
        {
            UI.PreviewInCollectionState(!IsInActiveCollection);
        }

        public void ResetToggleStatePreview()
        {
            RefreshUI();
        }

        public bool SetState(bool state)
        {
            IsInActiveCollection = state;
            RefreshUI();
            return state;
        }

        public bool SetCursor(bool state)
        {
            HasCursor = state;
            RefreshUI();
            return state;
        }

        public void ShowRemoveButton(bool doShow)
        {
            UI.RemoveButton.gameObject.SetActive(doShow);
        }

        public bool IsMouseOver()
        {
            return UI?.IsInitialised == true && UI.MouseOver.isOver;
        }
    }
}
