﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using UnityEngine;
using UnityEngine.Events;
using Object = UnityEngine.Object;

namespace AudioMate
{
    public class ClipCursor
    {
        public int Index;
        private AudioMateClip _clip;
        private readonly AudioMateController _c;

        public ClipCursor(AudioMateController controller)
        {
            _c = controller;
        }

        // Get Clip Cursor
        public AudioMateClip GetClip()
        {
            return _clip;
        }

        // Set Clip Cursor
        public void SetClip(AudioMateClip newCursorClip)
        {
            if ((Object) _c == (Object) null) return;
            Reset();
            _clip = newCursorClip;
            Index = _c.Clips.IndexOf(_clip);

            _clip.SetCursor(true);
        }

        private void Reset()
        {
            _c.Clips.ForEach(clip =>
            {
                clip.SetCursor(false);
            });
        }
    }

    public class ClipLibrary : MonoBehaviour
    {
        public class ClipsRefreshedEvent : UnityEvent { }
        public readonly ClipsRefreshedEvent OnClipsRefreshed = new ClipsRefreshedEvent();

        private AudioMateController _c;

        private List<NamedAudioClip> _sourceClipList;
        private bool _processing;

        public List<AudioMateClip> Clips { get; private set; }

        public ClipCursor Cursor;
        public bool isInitialised;
        private bool _isBound;

        public void Init(AudioMateController controller)
        {
            if ((Object) controller == (Object) null) return;
            _c = controller;
            Cursor = new ClipCursor(_c);
            Clips = new List<AudioMateClip>();
        }

        public void Bind()
        {
            Log("Binding Clip Library");
            _c.collections.OnActiveCollectionSelected.AddListener(OnActiveCollectionSelected);
            _c.collections.OnActiveCollectionUpdated.AddListener(OnActiveCollectionUpdated);
            _c.audioFileManager.OnSceneFilesUpdated.AddListener(OnSourceClipsUpdated);
            _isBound = true;
            _processing = false;
            RefreshClipIndex();
        }

        public void SetCursorIndex(int index)
        {
            if (Clips.ElementAtOrDefault(index) != null) Cursor?.SetClip(Clips.ElementAtOrDefault(index));
        }

        public void RefreshClipIndex()
        {
            if (_isBound == false) return;
            Log("Refreshing Clip Index");
            try
            {
                SuperController.singleton.StartCoroutine(DeferredIndexing());
            }
            catch (Exception e)
            {
                SuperController.LogError($"AudioMate.{nameof(ClipLibrary)}.{nameof(RefreshClipIndex)}: {e}");
            }
        }

        private IEnumerator DeferredIndexing()
        {
            if (_processing)
            {
                yield break;
            }
            _processing = true;
            _sourceClipList = URLAudioClipManager.singleton.GetCategoryClips("web");
            // Convert new clips to AudioMateClips and add them to the list
            var clipCount = 0;
            if (_sourceClipList != null)
            {
                foreach (var sourceClip in _sourceClipList)
                {
                    var audioMateClip = Clips.Find(x => x.AudioClip == sourceClip);
                    if (audioMateClip != null)
                    {
                        continue;
                    }
                    audioMateClip = new AudioMateClip(sourceClip);
                    Clips.Add(audioMateClip);
                    clipCount++;
                }

                foreach (var audioMateClip in Clips)
                {
                    if (audioMateClip.AudioClip == null && !audioMateClip.IsProvisional)
                    {
                        audioMateClip.Destroy();
                    }
                    var sourceClipExists = _sourceClipList.Exists(sourceClip => sourceClip == audioMateClip.AudioClip);
                    if (!sourceClipExists && !audioMateClip.IsProvisional) audioMateClip.Destroy();
                }
            }
            Log($"Added {clipCount} Clip(s) to library.");
            yield return null;
            ProcessProvisionalClips();
            yield return null;
            OnClipsRefreshed.Invoke();
            yield return null;
            _processing = false;
            yield return null;
            isInitialised = true;
        }

        public void RemoveAllClips()
        {
            _c.collections.ClearAllCollections();
            URLAudioClipManager.singleton.RemoveAllClips();
            _c.audioFileManager.OnSceneFilesUpdated.Invoke();
        }

        public void RemoveUnusedClips()
        {
            var unusedClips = _c.collections.UnusedClips();
            unusedClips.ForEach(clip =>
            {
                URLAudioClipManager.singleton.RemoveClip(clip.AudioClip);
            });
            _c.audioFileManager.OnSceneFilesUpdated.Invoke();
        }

        public AudioMateClip FindByAudioClipUID(string audioClipUID)
        {
            return Clips.Find(x => x.AudioClip.uid == audioClipUID);
        }

        public void ProcessProvisionalClips()
        {
            foreach (var collection in _c.collections.CollectionsWithProvisionalClips)
            {
                var isActive = collection == _c.collections.ActiveCollection;
                if (!collection.GetProvisionalClips().Any())
                {
                    continue;
                }
                foreach (var provisionalClip in collection.GetProvisionalClips())
                {
                    if (provisionalClip?.AudioClip == null || !provisionalClip.IsProvisional)
                    {
                        continue;
                    }

                    var audioMateClip = FindByAudioClipUID(provisionalClip.AudioClip.uid);
                    if (audioMateClip == null)
                    {
                        continue;
                    }

                    if (isActive) audioMateClip.SetState(true);
                    collection.SwapProvisionalClip(provisionalClip, audioMateClip);
                }
            }

            if (_c.collections.CollectionsWithProvisionalClips.Any())
            {
                _c.needsProvisionalClipProcessing = true;
            }
        }

        public void OnActiveCollectionSelected(CollectionManager.ActiveCollectionSelectedEventArgs args)
        {
            // ...
        }

        public void OnActiveCollectionUpdated()
        {
            // ...
        }

        public void OnSourceClipsUpdated()
        {
            Log("Scene Clips updated!");
            RefreshClipIndex();
        }

        public void OnDestroy()
        {
            try
            {
                _isBound = false;
                _processing = false;
                if (_c == null || _c.collections == null) return;
                _c.collections.OnActiveCollectionSelected.RemoveListener(OnActiveCollectionSelected);
                _c.collections.OnActiveCollectionUpdated.RemoveListener(OnActiveCollectionUpdated);
                _c.audioFileManager.OnSceneFilesUpdated.RemoveListener(OnSourceClipsUpdated);
            }
            catch (Exception e)
            {
                SuperController.LogMessage($"AudioMate: [UIClipLibrary.OnDestroy] {e}");
            }
        }

        private void Log(string message)
        {
            if (_c == null) return;
            _c.Log($" [ClipLibrary]: {message}");
        }
    }
}
