﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Globalization;
using System.Linq;
using SimpleJSON;
using UnityEngine;
using UnityEngine.Events;
using Object = UnityEngine.Object;

namespace AudioMate
{
    public class Collection : TriggerHandler, IEnumerable<AudioMateClip>
    {
        public class ClipPlayedEvent : UnityEvent {}
        public readonly ClipPlayedEvent OnClipPlayed = new ClipPlayedEvent();

        public string Name;
        public bool Enabled = true;
        public bool ShuffleMode = true;
        public bool AlwaysQueue;
        public bool OnlyIfClear;
        public bool OnlyIfThisCollectionIsNotPlaying;
        public bool AutoPlay;
        public bool HasAutoPlayBeenTriggered;
        public bool ContinuousPlay;
        public bool HasPlayedAnything;

        public string ReceiverAtomUID;
        public string ReceiverNodeUID;

        public Atom ReceiverAtom => ReceiverAtomUID != null ? SuperController.singleton.GetAtomByUid(ReceiverAtomUID) : null;
        public AudioSourceControl ReceiverAudioSource;

        public bool VolumeControlEnabled;
        public bool PitchControlEnabled;
        public float Volume = 1f;
        public float Pitch = 1f;
        public float PlayChance = 1f;
        public int Layer;
        public bool IgnoreLayers;
        public bool Loop = true;

        private float _volumeTriggerMultiplier = 15f;
        private float _volumeTriggerQuickness = 5f;

        private Dictionary<string, CollectionTrigger> _collisionTriggers;
        private CollectionTrigger _volumeTrigger;

        private AudioMateController _c;
        private int _currentClipIndex = -1;

        private readonly List<int> _unplayedClipsList = new List<int>();

        public IEnumerable<NamedAudioClip> NamedAudioClips
        {
            get
            {
                return Enumerable.ToList(Clips.Select(clip => clip.AudioClip));
            }
        }

        public List<AudioMateClip> Clips { get; } = new List<AudioMateClip>();

        public float VolumeTriggerMultiplier
        {
            get
            {
                return _volumeTriggerMultiplier != 0f ? _volumeTriggerMultiplier : 15f;
            }
            set
            {
                _volumeTriggerMultiplier = value;
                GetVolumeTrigger().SetVolumeTriggerMultiplier(value);
            }
        }

        public float VolumeTriggerQuickness
        {
            get
            {
                return _volumeTriggerQuickness != 0f ? _volumeTriggerQuickness : 5f;
            }
            set
            {
                _volumeTriggerQuickness = value;
                GetVolumeTrigger().SetVolumeTriggerQuickness(value);
            }
        }

        public int Count => Clips.Count;

        public bool Contains(AudioMateClip clip) => Clips.Contains(clip);
        public AudioMateClip Find(System.Predicate<AudioMateClip> predicate) => Clips.Find(predicate);

        public AudioMateClip CurrentClip
        {
           get {
                if (_currentClipIndex == -1) return null;
                return Clips.Count-1 >= _currentClipIndex ? Clips[_currentClipIndex] : null;
           }
        }

        public Collection(AudioMateController controller, string collectionName)
        {
            _c = controller;
            Name = collectionName;
            Init();
        }

        public void Init()
        {
            if ((Object) _c == (Object) null)
            {
                SuperController.LogError("Can not initialize Collection. Controller is null.");
            }
            ResetUnplayedClips();
            SyncAudioReceiver();
        }

        public virtual void OnAtomRename(string oldName, string newName)
        {
            ReceiverAtomUID = newName;
            SyncAudioReceiver();
            GetVolumeTrigger()?.SyncAtomNames();
            foreach (var collectionTrigger in GetCollisionTriggers()
                         .Where(collectionTrigger => collectionTrigger.Value != null))
            {
                collectionTrigger.Value.SyncAtomNames();
            }
        }

        public void ResetCollisionTriggers()
        {
            if (_collisionTriggers == null) {
                _collisionTriggers = new Dictionary<string, CollectionTrigger>();
            }
            else
            {
                foreach (var keyValuePair in _collisionTriggers)
                {
                    keyValuePair.Value?.Remove();
                }

                _collisionTriggers.Clear();
            }
        }

        public Dictionary<string, CollectionTrigger>GetCollisionTriggers()
        {
            if (_collisionTriggers == null) ResetCollisionTriggers();
            return _collisionTriggers;
        }

        public CollectionTrigger GetCollisionTrigger(string triggerUID)
        {
            if (_collisionTriggers == null)
            {
                ResetCollisionTriggers();
            }

            if (_collisionTriggers?.ContainsKey(triggerUID) != true)
            {
                // Create a new collision trigger for the given trigger uid, if it doesn't exist yet.
                var newCollisionTrigger = new CollectionTrigger(_c, TriggerCategories.Collision, triggerUID)
                {
                    displayName = triggerUID,
                    Collection = this,
                    handler = this
                };
                newCollisionTrigger.InitTriggerUI();
                newCollisionTrigger.InitTriggerActionsUI();
                _collisionTriggers?.Add(triggerUID, newCollisionTrigger);
            }

            var collisionTrigger = _collisionTriggers?[triggerUID];
            if (collisionTrigger != null)
            {
                collisionTrigger.Validate();
                return _collisionTriggers?[triggerUID];
            }
            SuperController.LogError($"Could not initialize Collision Trigger for UID {triggerUID} assigned to Collection {this.Name}");
            return null;
        }

        public void SetCollisionTriggerActiveState(string triggerUID, bool isActive)
        {
            GetCollisionTrigger(triggerUID).active = isActive;
        }

        public void ResetVolumeTrigger()
        {
            if (_volumeTrigger != null) _volumeTrigger.Remove();
            _volumeTrigger = null;
            _volumeTrigger = new CollectionTrigger(_c, TriggerCategories.Volume)
            {
                displayName = "Volume Trigger",
                Collection = this
            };
            _volumeTrigger.InitTriggerUI();
            _volumeTrigger.InitTriggerActionsUI();
            _volumeTrigger.handler = this;
        }

        public CollectionTrigger GetVolumeTrigger()
        {
            if (_volumeTrigger == null)
            {
                ResetVolumeTrigger();
            }

            if (_volumeTrigger != null)
            {
                if (_volumeTrigger.Collection == null)
                {
                    _volumeTrigger.Collection = this;
                }
                _volumeTrigger.Validate();
                return _volumeTrigger;
            }
            SuperController.LogError($"Could not initialize Volume Trigger assigned to Collection {this.Name}");
            return null;
        }

        public bool IsPlaying()
        {
            // TODO seems to do not work sometimes
            var currentClip = CurrentClip;
            if (currentClip == null) return false;
            return ReceiverAudioSource.audioSource.clip == currentClip.AudioClip.sourceClip && ReceiverAudioSource.audioSource.isPlaying;
        }

        public JSONClass GetJSON(bool excludeTriggerActions = false)
        {
            var clipsJSON = new JSONArray();
            if (Clips.Count > 0)
            {
                foreach (var clip in Clips)
                {
                    clipsJSON.Add(clip.GetJSON());
                }
            }

            var collectionJSON = new JSONClass
            {
                { "name", Name },
                { "enabled", Enabled.ToString() },
                { "receiverAtom", ReceiverAtomUID },
                { "receiverNode", ReceiverNodeUID },
                { "shuffle", ShuffleMode.ToString() },
                { "loop", Loop.ToString() },
                { "alwaysQueue", AlwaysQueue.ToString() },
                { "layer", Layer.ToString() },
                { "ignoreLayers", IgnoreLayers.ToString() },
                { "onlyIfClear", OnlyIfClear.ToString() },
                { "onlyIfThisCollectionIsNotPlaying", OnlyIfThisCollectionIsNotPlaying.ToString() },
                { "autoPlay", AutoPlay.ToString() },
                { "continuousPlay", ContinuousPlay.ToString() },
                { "playChance", PlayChance.ToString(CultureInfo.InvariantCulture) },
                { "pitch", Pitch.ToString(CultureInfo.InvariantCulture) },
                { "volume", Volume.ToString(CultureInfo.InvariantCulture) },
                { "lastClipIndex", _currentClipIndex.ToString() }, // Deprecated legacy value
                { "currentClipIndex", _currentClipIndex.ToString() },
                { "clips", clipsJSON },
                { "volumeTriggerMultiplier", _volumeTriggerMultiplier.ToString(CultureInfo.InvariantCulture) },
                { "volumeTriggerQuickness", _volumeTriggerQuickness.ToString(CultureInfo.InvariantCulture) },
            };

            InsertVariables(collectionJSON);

            if (excludeTriggerActions) return collectionJSON;
            var volumeTrigger = GetVolumeTrigger();
            if (volumeTrigger.VolumeTriggeredMorphActions().Count > 0)
            {
                collectionJSON.Add(Storables.VolumeTrigger, volumeTrigger.GetJSON());
            }

            var collisionTriggers = GetCollisionTriggers();
            var collisionTriggersJSON = new JSONClass();
            foreach (var keyValuePair in collisionTriggers)
            {
                var triggerUID = keyValuePair.Key;
                var collisionTrigger = keyValuePair.Value;
                if (collisionTrigger.HasTriggerActions())
                {
                    collisionTriggersJSON.Add(triggerUID, collisionTrigger.GetJSON());
                }

                if (collisionTriggersJSON.Childs.Any())
                {
                    collectionJSON.Add(Storables.CollisionTriggers, collisionTriggersJSON);
                }
            }
            return collectionJSON;
        }

        public bool RestoreFromJSON(JSONClass json, bool merge = false, bool doNotInit = false)
        {
            try
            {
                if (json?.AsObject == null || !json.AsObject.Childs.Any())
                {
                    SuperController.LogError($"AudioMate.{nameof(Collection)}.{nameof(RestoreFromJSON)} Can not parse collection data. JSON data is empty.");
                    return false;
                }

                if (!merge) Clips.Clear();

                ParseVariables(json);

                Name = json["name"].Value;
                Enabled = json["enabled"].AsBool;
                ReceiverAtomUID = json["receiverAtom"].Value;
                ReceiverNodeUID = json["receiverNode"].Value;
                ShuffleMode = json["shuffle"].AsBool;
                Loop = json["loop"] == null || json["loop"].AsBool;
                AlwaysQueue = json["alwaysQueue"].AsBool;
                Layer = json["layer"] != null ? json["layer"].AsInt : 0;
                IgnoreLayers = json["ignoreLayers"].AsBool;
                OnlyIfClear = json["onlyIfClear"].AsBool;
                OnlyIfThisCollectionIsNotPlaying = json["onlyIfThisCollectionIsNotPlaying"].AsBool;
                PlayChance = json["playChance"].AsFloat;
                AutoPlay = json["autoPlay"].AsBool;
                ContinuousPlay = json["continuousPlay"].AsBool;
                Pitch = json["pitch"] != null ? json["pitch"].AsFloat : 1f;
                Volume = json["volume"] != null ? json["volume"].AsFloat : 1f;
                VolumeTriggerMultiplier = json["volumeTriggerMultiplier"] != null ? json["volumeTriggerMultiplier"].AsFloat : 15f;
                VolumeTriggerQuickness = json["volumeTriggerQuickness"] != null ? json["volumeTriggerQuickness"].AsFloat : 5f;
                SyncAudioReceiver();

                _currentClipIndex = merge ? -1 :
                    json["currentClipIndex"] != null ? json["currentClipIndex"].AsInt :
                    json["lastClipIndex"] != null ? json["lastClipIndex"].AsInt :
                    -1;

                if (json["clips"] != null)
                {
                    foreach (JSONClass clip in json["clips"].AsArray)
                    {
                        if (HasClip(clip)) continue;

                        // Check for legacy formats of .amc file and update if possible
                        var lastFourCharacters = clip["uid"].ToString().Substring(Math.Max(0, clip["uid"].ToString().Length - 4));
                        if (lastFourCharacters.Equals(".wav", StringComparison.OrdinalIgnoreCase))
                        {
                            clip["uid"] = clip["uid"].ToString().Substring(0, Math.Max(0, clip["uid"].ToString().Length - 4));
                        }
                        if (!string.IsNullOrEmpty(clip["sourceClip"].Value) && string.IsNullOrEmpty(clip["audioClip"].Value))
                        {
                            clip["audioClip"].Value = clip["sourceClip"].Value;
                        }
                        var provisionalClip = new AudioMateClip(clip) { IsProvisional = true };
                        if (clip["displayName"] != null) provisionalClip.ProvisionalDisplayName = clip["displayName"];
                        Clips.Add(provisionalClip);
                    }
                }

                // Collision Triggers
                if (json[Storables.CollisionTriggers] != null)
                {
                    var asObject = json[Storables.CollisionTriggers].AsObject;
                    foreach (var triggerUID in asObject.Keys)
                    {
                        GetCollisionTrigger(triggerUID).RestoreFromJSON(asObject[triggerUID].AsObject);
                    }
                }

                // Volume Trigger
                if (json[Storables.VolumeTrigger] != null)
                {
                    GetVolumeTrigger().RestoreFromJSON(json[Storables.VolumeTrigger].AsObject);
                }

                ResetUnplayedClips();
                if (!doNotInit) Init();
            }
            catch (Exception e)
            {
                SuperController.LogError($"AudioMate.{nameof(Collection)}.{nameof(RestoreFromJSON)} {e}");
                return false;
            }
            return true;
        }

        public virtual void InsertVariables(JSONClass json)
        {
            if (json == null) throw new ArgumentNullException(nameof(json));
            if (!json.Childs.Any()) return;
            var pluginStoreId = _c.storeId;
            var containingAtomUID = _c.containingAtom.uid;

            if (json["receiverAtom"].Value == containingAtomUID)
            {
                json["receiverAtom"] = Variables.Person;
            }
            if (json["receiverNodeTarget"].Value == pluginStoreId)
            {
                json["receiverNodeTarget"] = Variables.Plugin;
            }
        }

        public virtual void ParseVariables(JSONClass json)
        {
            if (json == null) throw new ArgumentNullException(nameof(json));
            if (!json.Childs.Any()) return;
            var pluginStoreId = _c.storeId;
            var containingAtomUID = _c.containingAtom.uid;

            if (json["receiverAtom"].Value == Variables.Person)
            {
                json["receiverAtom"] = containingAtomUID;
            }
            if (json["receiverNode"].Value == Variables.Plugin)
            {
                json["receiverNode"] = pluginStoreId;
            }

        }

        /**
         * Adds a clip to this collections
         */
        public void Add(AudioMateClip clip)
        {
            if (Clips == null) return;
            if (Clips.Contains(clip)) return;
            Clips.Add(clip);
            ResetUnplayedClips();
        }

        /**
         * Removes a clip from this collection
         */
        public bool Remove(NamedAudioClip sourceClip)
        {
            if (Clips == null) return false;
            var clip = FindByNamedAudioClip(sourceClip);
            if (clip == null) return false;
            ResetUnplayedClips();
            return true;
        }

        /**
         * Removes a clip from this collection
         */
        public bool Remove(AudioMateClip clip)
        {
            if (Clips == null) return false;
            var result = Clips.Remove(clip);
            if (result) ResetUnplayedClips();
            return result;
        }

        public void Clear()
        {
            Clips.Clear();
            ResetUnplayedClips();
        }

        public bool HasClip(AudioMateClip clip)
        {
            return HasClip(clip.GetJSON());
        }

        public bool HasClip(JSONClass clip)
        {
            return Clips.Exists(x => clip["sourceClip"] == x.AudioClip);
        }

        public void SwapProvisionalClip(AudioMateClip provisionalClip, AudioMateClip audioMateClip)
        {
            var index = Clips.IndexOf(provisionalClip);
            Clips.Remove(provisionalClip);
            Clips.Insert(index, audioMateClip);
            provisionalClip.Destroy();
        }

        public void Enable()
        {
            Enabled = true;
        }

        public void Disable()
        {
            Enabled = false;
        }

        public void Update()
        {
            if (_collisionTriggers != null) {
                foreach (var keyValuePair in _collisionTriggers)
                {
                    keyValuePair.Value?.Update();
                }
            }

            _volumeTrigger?.Update();
        }

        public List<AudioMateClip> GetProvisionalClips()
        {
            return Clips.FindAll(x => x.IsProvisional);
        }

        /**
         * Checks if a NamedAudioClip already exists in this collection.
         */
        public bool ContainsNamedAudioClip(NamedAudioClip namedAudioClip)
        {
            return ContainsNamedAudioClip(namedAudioClip.uid);
        }

        private bool ContainsNamedAudioClip(string namedAudioClipUID)
        {
            return Clips != null && Clips.Exists(x => x.AudioClip.uid == namedAudioClipUID);
        }

        private AudioMateClip FindByNamedAudioClip(NamedAudioClip sourceClip)
        {
            return Clips?.Find(x => x.AudioClip == sourceClip);
        }

        /**
         * Evaluate play chance and if it succeeds get the next clip after the last played one from the clip collection.
         * If no clip has been played yet, return the first clip.
         */
        private AudioMateClip GetNextClip(bool skipPlayChance = false)
        {
            if (Clips == null) return null;
            var nextClip = (AudioMateClip) null;

            try
            {
                // PlayChance test
                var randomChance = skipPlayChance ? 1f : (float) UnityEngine.Random.Range(0, 100) / 100;

                if (!skipPlayChance && PlayChance < randomChance) return null;
                var nextClipIndex = _currentClipIndex + 1;
                if (nextClipIndex > Clips.Count - 1)
                {
                    if (Loop)
                    {
                        nextClipIndex = 0;
                    }
                    else
                    {
                        return null;
                    }
                }
                nextClip = Clips.ElementAtOrDefault(nextClipIndex);
            }
            catch (Exception e)
            {
                SuperController.LogError($"AudioMate.{nameof(Collection)}.{nameof(GetNextClip)}: {e}");
            }

            return nextClip;
        }

        private AudioMateClip GetPreviousClip(bool skipPlayChance = false)
        {
            if (Clips == null) return null;
            var previousClip = (AudioMateClip) null;

            try
            {
                // PlayChance test
                var randomChance = skipPlayChance ? 1f : (float) UnityEngine.Random.Range(0, 100) / 100;

                if (!skipPlayChance && PlayChance < randomChance) return null;
                var previousClipIndex = _currentClipIndex - 1;
                if (previousClipIndex < 0)
                {
                    if (Loop)
                    {
                        previousClipIndex = Clips.Count-1;
                    }
                    else
                    {
                        return null;
                    }
                }
                previousClip = Clips.ElementAtOrDefault(previousClipIndex);
            }
            catch (Exception e)
            {
                SuperController.LogError($"AudioMate.{nameof(Collection)}.{nameof(GetPreviousClip)}: {e}");
            }

            return previousClip;
        }

        private AudioMateClip GetCurrentClip(bool skipPlayChance = false)
        {
            if (Clips == null) return null;
            var currentClip = (AudioMateClip) null;

            try
            {
                // PlayChance test
                var randomChance = skipPlayChance ? 1f : (float) UnityEngine.Random.Range(0, 100) / 100;

                if (!skipPlayChance && PlayChance < randomChance) return null;
                currentClip = Clips.ElementAtOrDefault(_currentClipIndex);
            }
            catch (Exception e)
            {
                SuperController.LogError($"AudioMate.{nameof(Collection)}.{nameof(GetCurrentClip)}: {e}");
            }

            return currentClip;
        }

        /**
         * Evaluate play chance and if it succeeds get a randomly (shuffle) selected clip from the clip collection.
         */
        private AudioMateClip GetRandomClip(bool skipPlayChance = false)
        {
            if (Clips == null) return null;
            var randomIndex = -1;
            try
            {
                // PlayChance test
                var randomChance = (float) UnityEngine.Random.Range(0, 100) / 100;

                if (!skipPlayChance && PlayChance < randomChance) return null;

                // ShuffleMode play
                if (_unplayedClipsList.Count == 0)
                {
                    if (Loop)
                    {
                        ResetUnplayedClips();
                    }
                    else
                    {
                        return null;
                    }
                }
                randomIndex = UnityEngine.Random.Range(0, _unplayedClipsList.Count); // max value is exclusive, that's why I'm not removing 1 from _unplayedClipsList.Count
                var randomClipIndex = _unplayedClipsList.ElementAtOrDefault(randomIndex);

                return Clips.ElementAtOrDefault(randomClipIndex);
            }
            catch (Exception e)
            {
                SuperController.LogError($"AudioMate.{nameof(Collection)}.{nameof(GetRandomClip)}: randomIndex: {randomIndex} {e}");
            }

            return null;
        }

        /**
         * Set the audio source which plays the clips
         */
        public void SyncAudioReceiver()
        {
            if (ReceiverAtomUID == null)
            {
                ReceiverAtomUID = _c.containingAtom.uid;
            }
            if (ReceiverNodeUID == null)
            {
                ReceiverNodeUID = AudioMateController.GuessInitialReceivingNode(ReceiverAtom);
            }

            if (ReceiverAtomUID == null || ReceiverNodeUID == null)
            {
                SuperController.LogError("AudioMate.Collection.SyncAudioReceiver: Can't sync Audio Receiver because Receiver has not been set and can't be guessed.");
                return;
            }
            try
            {
                var rA = SuperController.singleton.GetAtomByUid(ReceiverAtomUID);
                if ((Object) rA == (Object) null) return;
                ReceiverAudioSource =
                    rA.GetStorableByID(ReceiverNodeUID)
                        as AudioSourceControl;
            }
            catch (Exception e)
            {
                SuperController.LogError($"AudioMate.{nameof(Collection)}.{nameof(SyncAudioReceiver)}: {e}");
            }
        }

        private static class PlayBehaviour
        {
            public const string Random = "random";
            public const string Next = "next";
            public const string Previous = "previous";
            public const string Current = "current";
        }

        /**
		* Plays or queues a clip from the collection. If shuffle mode is enabled, a random clip will be selected.
        * Also takes layer settings into account and wont play the clip, if a clip on a higher layer is
        * currently playing, unless "Ignore Layers" is enabled.
		*/
        public void PlayClip(string forceBehaviour = null, bool forceQueue = false) {
            try
            {
                if ((Object) ReceiverAudioSource == (Object) null)
                {
                    SuperController.LogMessage("WARNING: the triggered collection has an invalid receiver.");
                }
                if (!Enabled || Clips.Count == 0) return;
                if (!IgnoreLayers && _c.collections.GetPlayingCollectionsAboveLayer(Layer, ReceiverAudioSource).Count > 0) return;

                AudioMateClip nextClip;
                switch (forceBehaviour)
                {
                    case PlayBehaviour.Random:
                        nextClip = GetRandomClip();
                        break;
                    case PlayBehaviour.Next:
                        nextClip = GetNextClip();
                        break;
                    case PlayBehaviour.Previous:
                        nextClip = GetPreviousClip();
                        break;
                    case PlayBehaviour.Current:
                        nextClip = GetCurrentClip();
                        break;
                    default:
                        nextClip = ShuffleMode ? GetRandomClip() : GetNextClip();
                        break;
                }
                if (nextClip == null) return;

                if (VolumeControlEnabled) ReceiverAudioSource.volume = Volume;
                if (PitchControlEnabled) ReceiverAudioSource.pitch = Pitch;
                if (AlwaysQueue || forceQueue)
                {
                    nextClip.Queue(ReceiverAudioSource);
                    _currentClipIndex = Clips.IndexOf(nextClip);
                    _unplayedClipsList.Remove(_currentClipIndex);
                    OnClipPlayed.Invoke();
                } else
                {
                    if (OnlyIfThisCollectionIsNotPlaying && IsPlaying())
                    {
                        return;
                    }
                    if (!nextClip.Play(ReceiverAudioSource, OnlyIfClear)) return;
                    HasPlayedAnything = true;
                    _currentClipIndex = Clips.IndexOf(nextClip);
                    _unplayedClipsList.Remove(_currentClipIndex);
                    OnClipPlayed.Invoke();
                }
            }
            catch (Exception e)
            {
                SuperController.LogError($"AudioMate.{nameof(Collection)}.{nameof(PlayRandomClip)}: {e}");
            }
        }

        public void QueueClip() {
            try {
                if (!Enabled) return;
                if (Clips.Count < 1 || (Object) ReceiverAudioSource == (Object) null) return;

                var nextClip = ShuffleMode ? GetRandomClip() : GetNextClip();
                if (nextClip == null) return;
                if (VolumeControlEnabled) ReceiverAudioSource.volume = Volume;
                if (PitchControlEnabled) ReceiverAudioSource.pitch = Pitch;
                nextClip.Queue(ReceiverAudioSource);
                _currentClipIndex = Clips.IndexOf(nextClip);
                OnClipPlayed.Invoke();
            }
            catch (Exception e)
            {
                SuperController.LogError($"AudioMate.{nameof(Collection)}.{nameof(QueueClip)}: {e}");
            }
        }

        /**
		* Plays a randomly selected clip.
		*/
        public void PlayRandomClip() {
            PlayClip(PlayBehaviour.Random);
        }

        /**
		* Queues a randomly selected clip.
		*/
        public void QueueRandomClip() {
            PlayClip(PlayBehaviour.Random, true);
        }

        /**
		* Plays the clip after the current clip index.
		*/
        public void PlayNextClip(bool queue = false) {
            PlayClip(PlayBehaviour.Next, queue);
        }

        /**
		* Plays the clip before the current clip index.
		*/
        public void PlayPreviousClip(bool queue = false) {
            PlayClip(PlayBehaviour.Previous, queue);
        }

        /**
		* Plays the clip at the current clip index.
		*/
        public void PlayCurrentClip(bool queue = false) {
            PlayClip(PlayBehaviour.Current, queue);
        }

        /**
		* Sets the clip index and the cursor. Resets the unplayed clips list.
		*/
        public void SetPlayIndex(int playIndex)
        {
            if (Clips?[playIndex] == null) return;
            _currentClipIndex = playIndex;
            Clips[playIndex].SetCursor(true);
            ResetUnplayedClips();
        }

        /**
		* Sets the clip index and the cursor to 0. Resets the unplayed clips list.
		*/
        public void ResetPlayIndex()
        {
            if (Clips != null && Clips.Count > 0)
            {
                SetPlayIndex(0);
            }
        }

        /**
         * Reset the unplayed clips list, which is used for the shuffle mode. The list is a simple integer index list,
         * representing the indexes of all clips in the collection. Whenever a clip gets played, its also being removed
         * from the unplayed clips list. The randomizer can only select clip indexes from that list.
         */
        private void ResetUnplayedClips()
        {
            _unplayedClipsList.Clear();
            if (Clips == null) return;
            for (var i = 0; i < Clips.Count; i++)
            {
                _unplayedClipsList.Add(i);
            }
        }

        public void Destroy()
        {
            Clips?.Clear();

            if (_collisionTriggers != null) {
                foreach (var keyValuePair in _collisionTriggers)
                {
                    var trigger = keyValuePair.Value;
                    trigger?.Remove();
                }
                _collisionTriggers.Clear();
            }

            _volumeTrigger?.Remove();
            _volumeTrigger = null;
        }

        public override string ToString()
        {
            return Clips.Aggregate("", (current, clip) => current + (clip.DisplayName + "\n"));
        }

        IEnumerator<AudioMateClip> IEnumerable<AudioMateClip>.GetEnumerator()
        {
            return Clips.GetEnumerator();
        }

        public IEnumerator GetEnumerator()
        {
            return Clips.GetEnumerator();
        }

        public void RemoveTrigger(Trigger t)
        {
            // nothing to do here
        }

        public void DuplicateTrigger(Trigger t)
        {
            // nothing to do here
        }

        public RectTransform CreateTriggerActionsUI()
        {
            return _c.ui.mainPanel.CreateTriggerActionsUI();
        }

        public RectTransform CreateTriggerActionDiscreteUI()
        {
            return _c.ui.mainPanel.CreateTriggerActionDiscreteUI();
        }

        public RectTransform CreateTriggerActionMiniUI()
        {
            return _c.ui.mainPanel.CreateTriggerActionMiniUI();
        }

        public RectTransform CreateTriggerActionTransitionUI()
        {
            return _c.ui.mainPanel.CreateTriggerActionTransitionUI();
        }

        public void RemoveTriggerActionUI(RectTransform rt)
        {
            _c.ui.mainPanel.RemoveTriggerActionUI(rt);
        }
    }
}
