﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using SimpleJSON;
using UnityEngine;
using UnityEngine.Events;
using Object = UnityEngine.Object;

namespace AudioMate
{
    public class CollectionManager : MonoBehaviour, IEnumerable<Collection>
    {

        private class CollectionActions
        {
            public JSONStorableAction EnableAction;                 // Enables the collection
            public JSONStorableAction DisableAction;                // Disables the collection
            public JSONStorableAction SetActiveAction;              // Sets the collection as active collection
            public JSONStorableAction PlayAction;                   // Plays the collection according to the user's collection settings
            public JSONStorableAction QueueAction;                  // Queues the collection according to the user's collection settings
            public JSONStorableAction PlayRandomAction;             // Always plays a random clip from the collection, ignores shuffle being set to false
            public JSONStorableAction QueueRandomAction;            // Always plays a random clip from the collection, ignores shuffle being set to false
            public JSONStorableAction PlayNextAction;               // Always plays the next clip, ignores shuffle being set to true
            public JSONStorableAction PlayPreviousAction;           // Always plays the previous clip, ignores shuffle being set to true
            public JSONStorableAction PlayCurrentAction;            // Always plays the clip at the current play index, ignores shuffle being set to true
            public JSONStorableAction ResetPlayIndexAction;         // Resets the play index to zero and resets unplayed clip list
        }

        public class ActiveCollectionSelectedEventArgs { public Collection Before; public Collection After; }
        public class ActiveCollectionNameChangedEventArgs { public string Before; public string After; }

        public class ActiveCollectionSelectedEvent : UnityEvent<ActiveCollectionSelectedEventArgs> { }
        public class ActiveCollectionUpdatedEvent : UnityEvent { }
        public class ActiveCollectionNameChangedEvent : UnityEvent<ActiveCollectionNameChangedEventArgs> { }

        public readonly ActiveCollectionSelectedEvent OnActiveCollectionSelected = new ActiveCollectionSelectedEvent();
        public readonly ActiveCollectionUpdatedEvent OnActiveCollectionUpdated = new ActiveCollectionUpdatedEvent();
        public readonly ActiveCollectionNameChangedEvent OnActiveCollectionNameChanged = new ActiveCollectionNameChangedEvent();

        private AudioMateController _c;
        private readonly List<Collection> _collections = new List<Collection>();

        public IEnumerator<Collection> GetEnumerator()
        {
            return _collections.GetEnumerator();
        }

        IEnumerator IEnumerable.GetEnumerator()
        {
            return GetEnumerator();
        }

        public IEnumerable<Collection> CollectionsWithProvisionalClips =>
            _collections.FindAll(x => x.GetProvisionalClips().Count > 0);

        private readonly Dictionary<string,CollectionActions> _collectionActions = new Dictionary<string, CollectionActions>();

        public JSONStorableStringChooser CollectionsJSON;

        public Collection ActiveCollection { get; private set; }

        public IEnumerable<string> Names => _collections.Select(collection => collection.Name).Distinct().ToList();
        public bool isInitialized;
        public bool isParsingCollections;
        public bool isAutoPlayRoutineRunning;
        private Coroutine _autoPlayRoutine;

        #region Initialization
        public void Init(AudioMateController controller)
        {
            if ((Object) controller == (Object) null)
            {
                SuperController.LogError($"AudioMate.{nameof(CollectionManager)}.{nameof(Init)}: Controller is null.");
                return;
            }
            _c = controller;
            if (isInitialized)
            {
                return;
            }
            Log("Initializing Collection Manager");
            if ((Object) _c.ui == (Object) null)
            {
                SuperController.LogError($"AudioMate.{nameof(CollectionManager)}.{nameof(Init)}: UI is not initialized.");
                return;
            }
            _c.OnContainingAtomNameChanged.AddListener(OnAtomRename);
            CollectionsJSON = _c.CollectionsJSON;
            CollectionsJSON.setCallbackFunction = SelectActiveCollection;
            SelectActiveCollection();
            isInitialized = true;
        }

        public void StartAutoPlayRoutine()
        {
            if (isAutoPlayRoutineRunning || _collections == null) return;
            if (_autoPlayRoutine != null) StopCoroutine(_autoPlayRoutine);
            _autoPlayRoutine = StartCoroutine(AutoPlayDeferred());
        }

        private IEnumerator AutoPlayDeferred()
        {
            while (true)
            {
                if (this == null) yield break;
                if (isAutoPlayRoutineRunning == false) isAutoPlayRoutineRunning = true;
                foreach (var audioMateClipCollection in _collections.FindAll(x => x.AutoPlay && !x.HasAutoPlayBeenTriggered))
                {
                    audioMateClipCollection.HasAutoPlayBeenTriggered = true;
                    audioMateClipCollection.PlayClip();
                }
                foreach (var audioMateClipCollection in _collections.FindAll(x => x.ContinuousPlay))
                {
                    if (!audioMateClipCollection.IsPlaying() && audioMateClipCollection.HasPlayedAnything) audioMateClipCollection.PlayClip();
                }

                yield return new WaitForSeconds(.2f);
            }
        }

        private void AddDefaultCollection()
        {
            if (_collections == null) return;

            var newCollection = AddCollection();
            SelectActiveCollection(newCollection);
        }
        #endregion

        #region Helper
        private string GetAvailableDefaultCollectionName()
        {
            var collectionName = GetAvailableCollectionName("Untitled");
            return collectionName;
        }

        public string GetAvailableCollectionName(string collectionName)
        {
            if (!Names.Contains(collectionName)) return collectionName;
            var newName = collectionName;
            var i = 2;
            while (Names.Contains(newName))
            {
                newName = collectionName + i;
                i++;
            }

            return newName;
        }

        public void SyncCollectionNames()
        {
            var choices = new List<string>();
            choices.AddRange(Names);
            CollectionsJSON.choices = choices;
        }
        #endregion

        #region Collection Management

        public Collection AddCollection(string collectionName = null)
        {
            if (string.IsNullOrEmpty(collectionName))
            {
                collectionName = GetAvailableDefaultCollectionName();
            }
            var newCollection = new Collection(_c, collectionName);
            return AddCollection(newCollection);
        }

        private Collection AddCollection(Collection collection)
        {
            if (collection == null || string.IsNullOrEmpty(collection.Name)) return null;
            if (collection.ReceiverAtomUID == null) collection.ReceiverAtomUID = _c.containingAtom.uid;
            if (collection.ReceiverNodeUID == null) collection.ReceiverNodeUID = AudioMateController.GuessInitialReceivingNode(_c.containingAtom);
            collection.OnClipPlayed.AddListener(() => OnClipPlayed(collection));
            collection.VolumeControlEnabled = _c.VolumeControlEnabledJSON.val;
            collection.PitchControlEnabled = _c.PitchControlEnabledJSON.val;
            var existingName = _collections.FindLastIndex(x => x.Name == collection.Name);
            if (existingName == -1)
            {
                _collections.Add(collection);
            }
            else
            {
                _collections.Insert(existingName, collection);
            }
            AddNewCollectionActions(collection.Name);
            if (!CollectionsJSON.choices.Contains(collection.Name))
            {
                CollectionsJSON.choices.Add(collection.Name);
            }
            SelectActiveCollection(collection);

            return collection;
        }

        private void OnClipPlayed(Collection collection)
        {
            var index = _c.Clips.IndexOf(collection.CurrentClip);
            _c.clipLibrary.SetCursorIndex(index);
        }

        public bool RemoveCollection(string collectionName)
        {
            if (string.IsNullOrEmpty(collectionName)) return false;
            var collectionToRemove = _collections.Find(x => x.Name == collectionName);
            return collectionToRemove != null && RemoveCollection(collectionToRemove);
        }

        private bool RemoveCollection(Collection collection)
        {
            if (collection == null) return false;
            var collectionName = collection.Name;
            if (string.IsNullOrEmpty(collectionName)) return false;
            CollectionsJSON.choices.Remove(collectionName);
            RemoveCollectionActions(collectionName);
            var result = _collections.Remove(collection);
            SyncCollectionNames();
            collection.Destroy();
            return result;
        }

        public Collection GetCollection(string collectionName)
        {
            return _collections.Find(x => x.Name == collectionName);
        }

        public bool RemoveActiveCollection()
        {
            var collectionIndex = _collections.IndexOf(ActiveCollection);
            var result = RemoveCollection(ActiveCollection);
            if (_collections.Count == 0)
            {
                AddDefaultCollection();
            }
            else
            {
                collectionIndex--;
                SelectActiveCollection(collectionIndex == -1 ? _collections.Last() : _collections[collectionIndex]);
            }
            return result;
        }

        public void SelectActiveCollection(Collection collection = null)
        {
            if (collection == null && ActiveCollection != null) return;
            if (collection != null && ActiveCollection == collection) return;
            while (true)
            {
                if (collection != null)
                {
                    if (collection.Name == ActiveCollection?.Name) break;
                    var previous = ActiveCollection;
                    ActiveCollection = collection;
                    ActiveCollection.SyncAudioReceiver();
                    if (CollectionsJSON.val != ActiveCollection.Name) CollectionsJSON.val = ActiveCollection.Name;
                    OnActiveCollectionSelected.Invoke(new ActiveCollectionSelectedEventArgs {Before = previous, After = ActiveCollection});
                    break;
                }

                if (_collections.Count > 0)
                {
                    collection = _collections.First();
                    continue;
                }

                AddDefaultCollection();
                collection = null;
            }
        }

        public void SelectActiveCollection(string collectionName)
        {
            if (string.IsNullOrEmpty(collectionName))
            {
                Log($"WARNING: Can't select Active Collection because Collection name is missing.");
                return;
            }
            var collection = _collections.Find(x => x.Name == collectionName);
            if (collection == null)
            {
                Log($"WARNING: Can't select Active Collection because Collection is not in Collection list.");
                return;
            }
            Log($"Selecting {collectionName} as Active Collection");
            SelectActiveCollection(collection);
        }

        public void Enable(string collectionName)
        {
            var collection = _collections.Find(x => x.Name == collectionName);
            collection?.Enable();
        }

        public void Enable()
        {
            ActiveCollection?.Enable();
        }

        public void Disable(string collectionName)
        {
            var collection = _collections.Find(x => x.Name == collectionName);
            collection?.Disable();
        }

        public void Disable()
        {
            ActiveCollection?.Disable();
        }

        public void EnableAll()
        {
            foreach (var collection in _collections)
            {
                collection.Enable();
            }
        }

        public void DisableAll()
        {
            foreach (var collection in _collections)
            {
                collection.Disable();
            }
        }

        public void SetActiveCollectionName(string newCollectionName)
        {
            if (string.IsNullOrEmpty(newCollectionName)) return;
            if (ActiveCollection == null) return;
            var oldCollectionName = ActiveCollection.Name;
            ActiveCollection.Name = newCollectionName;
            SyncCollectionNames();
            CollectionsJSON.val = newCollectionName;
            UpdateCollectionActionNames(oldCollectionName, newCollectionName);

            OnActiveCollectionNameChanged.Invoke(new ActiveCollectionNameChangedEventArgs
            {
                Before = oldCollectionName,
                After = newCollectionName
            });
        }

        /**
         * Adds a clip to the active collections
         */
        public void AddClipToActiveCollection(AudioMateClip clip, bool noEvent = false)
        {
            ActiveCollection?.Add(clip);
            if (!noEvent) OnActiveCollectionUpdated.Invoke();
        }

        /**
         * Adds a range of clips to the active collections
         */
        public void AddClipRangeToActiveCollection(int fromIndex, int toIndex, bool noEvent = false)
        {
            if (ActiveCollection == null) return;
            for (var i = fromIndex; i <= toIndex; i++)
            {
                var clip = _c.clipLibrary.Clips.ElementAtOrDefault(i);
                if (clip == null) continue;
                ActiveCollection.Add(clip);
                _c.clipLibrary.SetCursorIndex(i);
            }
            if (!noEvent) OnActiveCollectionUpdated.Invoke();
        }

        /**
         * Removes a clip from a collection
         */
        public bool RemoveClipFromActiveCollection(AudioMateClip clip, bool noEvent = false)
        {
            if (ActiveCollection == null || !ActiveCollection.Remove(clip)) return false;
            if (!noEvent) OnActiveCollectionUpdated.Invoke();
            return true;

        }

        /**
         * Removes a clip from all collections
         */
        public bool RemoveClipFromAllCollections(AudioMateClip clip, bool noEvent = false)
        {
            if (ActiveCollection == null || !ActiveCollection.Remove(clip)) return false;
            foreach (var collection in _collections)
            {
                if (collection.HasClip(clip)) collection.Remove(clip);
            }
            if (!noEvent) OnActiveCollectionUpdated.Invoke();
            return true;

        }

        /**
         * Adds all available clips to the active collections
         */
        public void AddAllClipsToActiveCollection()
        {
            foreach (var clip in _c.clipLibrary.Clips)
            {
                AddClipToActiveCollection(clip, true);
            }

            OnActiveCollectionUpdated.Invoke();
        }

        /**
         * Adds all clips visible in the clip library to the active collections. This takes current filter settings into
         * account.
         */
        public void AddVisibleClipsToActiveCollection()
        {
            foreach (var clip in _c.clipLibrary.Clips.Where(clip => clip.IsVisible))
            {
                AddClipToActiveCollection(clip, true);
            }

            OnActiveCollectionUpdated.Invoke();
        }

        public void ToggleClipsInActiveCollection(int startIndex, int amount, bool noEvent = false)
        {
            if (ActiveCollection == null || amount == 0) return;
            for (var i = startIndex; i <= startIndex + amount - 1; i++)
            {
                var clip = _c.clipLibrary.Clips.ElementAtOrDefault(i);
                if (clip == null) continue;
                if (clip.IsInActiveCollection)
                {
                    ActiveCollection.Remove(clip);
                }
                else
                {
                    ActiveCollection.Add(clip);
                }

                clip.ToggleState();
                _c.clipLibrary.SetCursorIndex(i);
            }
            if (!noEvent) OnActiveCollectionUpdated.Invoke();
        }

        /**
         * Removes all clips from the active collections
         */
        public void ClearActiveCollection()
        {
            ActiveCollection?.Clear();
            OnActiveCollectionUpdated.Invoke();
        }

        public void ClearAllCollections()
        {
            _collections.ForEach(collection => collection.Clear());
            OnActiveCollectionUpdated.Invoke();
        }

        public void SetVolumeControl(bool val)
        {
            _collections.ForEach(collection => collection.VolumeControlEnabled = val);
        }

        public void SetPitchControl(bool val)
        {
            _collections.ForEach(collection => collection.PitchControlEnabled = val);
        }

        public List<Collection> GetPlayingCollectionsAboveLayer(int layer, AudioSourceControl receiver)
        {
            return _collections.FindAll(x => x.Layer > layer && x.ReceiverAudioSource == receiver && x.IsPlaying());
        }

        public List<Collection> GetPlayingCollectionsAboveLayer(string layer, AudioSourceControl receiver)
        {
            return layer == null ? new List<Collection>() : GetPlayingCollectionsAboveLayer(int.Parse(layer), receiver);
        }
        #endregion

        #region Clip Actions

        /**
         * Plays a clip from a specific Collection. If shuffle mode is enabled, a random clip will be played.
         */
        public void PlayClipFromCollection(string collectionName, bool queue = false)
        {
            var collection = _collections.Find(x => x.Name == collectionName);
            if (queue)
            {
                collection?.QueueClip();
            }
            else
            {
                collection?.PlayClip();
            }
        }

        public void PlayNextClipFromCollection(string collectionName, bool queue = false)
        {
            var collection = _collections.Find(x => x.Name == collectionName);
            collection?.PlayNextClip(queue);
        }

        public void PlayPreviousClipFromCollection(string collectionName, bool queue = false)
        {
            var collection = _collections.Find(x => x.Name == collectionName);
            collection?.PlayPreviousClip(queue);
        }

        public void PlayCurrentClipFromCollection(string collectionName, bool queue = false)
        {
            var collection = _collections.Find(x => x.Name == collectionName);
            collection?.PlayPreviousClip(queue);
        }

        public void SetPlayIndexOfCollection(int playIndex, string collectionName)
        {
            var collection = _collections.Find(x => x.Name == collectionName);
            collection?.SetPlayIndex(playIndex);
        }

        public void ResetPlayIndexOfCollection(string collectionName)
        {
            var collection = _collections.Find(x => x.Name == collectionName);
            collection?.ResetPlayIndex();
        }

        /**
         * Queues a clip from a specific Collection. If shuffle mode is enabled, a random clip will be queued.
         */
        public void QueueClipFromCollection(string collectionName)
        {
            PlayClipFromCollection(collectionName, true);
        }

        /**
         * Plays a random clip from a specific Collection
         */
        public void PlayRandomClipFromCollection(string collectionName, bool queue = false)
        {
            var collection = _collections.Find(x => x.Name == collectionName);
            if (queue)
            {
                collection?.QueueRandomClip();
            }
            else
            {
                collection?.PlayRandomClip();
            }
        }

        /**
         * Queues a random clip from a specific Collection
         */
        public void QueueRandomClipFromCollection(string collectionName)
        {
            PlayRandomClipFromCollection(collectionName, true);
        }

        // Actions
        public void PlayClipAction(string collectionName = null)
        {
            if (collectionName == null)
            {
                ActiveCollection?.PlayClip();
            }
            else
            {
                PlayClipFromCollection(collectionName);
            }
        }

        public void PlayNextClipAction(string collectionName = null)
        {
            if (collectionName == null)
            {
                ActiveCollection?.PlayNextClip();
            }
            else
            {
                PlayNextClipFromCollection(collectionName);
            }
        }

        public void PlayPreviousClipAction(string collectionName = null)
        {
            if (collectionName == null)
            {
                ActiveCollection?.PlayPreviousClip();
            }
            else
            {
                PlayPreviousClipFromCollection(collectionName);
            }
        }

        public void PlayCurrentClipAction(string collectionName = null)
        {
            if (collectionName == null)
            {
                ActiveCollection?.PlayCurrentClip();
            }
            else
            {
                PlayCurrentClipFromCollection(collectionName);
            }
        }

        public void SetPlayIndexAction(int playIndex, string collectionName = null)
        {
            if (collectionName == null)
            {
                ActiveCollection?.SetPlayIndex(playIndex);
            }
            else
            {
                SetPlayIndexOfCollection(playIndex, collectionName);
            }
        }

        public void ResetPlayIndexAction(string collectionName = null)
        {
            if (collectionName == null)
            {
                ActiveCollection?.ResetPlayIndex();
            }
            else
            {
                ResetPlayIndexOfCollection(collectionName);
            }
        }

        public void QueueClipAction(string collectionName = null)
        {
            if (collectionName == null)
            {
                ActiveCollection?.QueueClip();
            }
            else
            {
                QueueClipFromCollection(collectionName);
            }
        }

        public void PlayRandomClipAction(string collectionName = null)
        {
            if (collectionName == null)
            {
                ActiveCollection?.PlayRandomClip();
            }
            else
            {
                PlayRandomClipFromCollection(collectionName);
            }
        }

        public void QueueRandomClipAction(string collectionName = null)
        {
            if (collectionName == null)
            {
                ActiveCollection?.QueueRandomClip();
            }
            else
            {
                QueueRandomClipFromCollection(collectionName);
            }
        }

        public IEnumerable<AudioMateClip> UsedClips(string collectionFilter = null)
        {
            var result = new List<AudioMateClip>();

            if (collectionFilter == null)
            {
                foreach (var collection in _collections)
                {
                    result.AddRange(collection.Clips);
                }
            }
            else
            {
                var collection = _collections.Find(x => x.Name == collectionFilter);
                if (collection == null) return result;
                result = collection.Clips;
            }

            return result;
        }

        public List<AudioMateClip> UnusedClips(string collectionFilter = null)
        {
            var result = new List<AudioMateClip>();
            result.AddRange(_c.Clips);
            if (collectionFilter == null)
            {
                foreach (var clip in _collections.SelectMany(collection => collection.Cast<AudioMateClip>()))
                {
                    result.Remove(clip);
                }
            }
            else
            {
                var collection = _collections.Find(x => x.Name == collectionFilter);
                if (collection == null) return result;
                foreach (AudioMateClip clip in collection)
                {
                    result.Remove(clip);
                }
            }
            return result;
        }
        #endregion

        /**
         * Creates all needed Actions for a new collection
         */
        public void AddNewCollectionActions(string collectionName)
        {
            if (string.IsNullOrEmpty(collectionName))
            {
                return;
            }
            if (_collectionActions == null) return;
            if (_collectionActions.ContainsKey(collectionName)) return;
            var newActions = new CollectionActions
            {
                EnableAction = new JSONStorableAction(GetEnableCollectionActionName(collectionName),
                    () => Enable(collectionName)),
                DisableAction = new JSONStorableAction(GetDisableCollectionActionName(collectionName),
                    () => Disable(collectionName)),
                SetActiveAction = new JSONStorableAction(GetSetActiveCollectionActionName(collectionName),
                    () => SelectActiveCollection(collectionName)),
                PlayAction = new JSONStorableAction(GetPlayClipActionName(collectionName),
                    () => PlayClipAction(collectionName)),
                QueueAction = new JSONStorableAction(GetQueueClipActionName(collectionName),
                    () => QueueClipAction(collectionName)),
                PlayRandomAction = new JSONStorableAction(GetPlayRandomClipActionName(collectionName),
                    () => PlayClipAction(collectionName)),
                QueueRandomAction = new JSONStorableAction(GetQueueRandomClipActionName(collectionName),
                    () => QueueClipAction(collectionName)),
                PlayNextAction = new JSONStorableAction(GetPlayNextClipActionName(collectionName),
                    () => PlayNextClipAction(collectionName)),
                PlayPreviousAction = new JSONStorableAction(GetPlayPreviousClipActionName(collectionName),
                    () => PlayPreviousClipAction(collectionName)),
                PlayCurrentAction = new JSONStorableAction(GetPlayCurrentClipActionName(collectionName),
                    () => PlayCurrentClipAction(collectionName)),
                ResetPlayIndexAction = new JSONStorableAction(GetResetPlayIndexActionName(collectionName),
                    () => ResetPlayIndexAction(collectionName)),
            };

            _c.RegisterAction(newActions.EnableAction);
            _c.RegisterAction(newActions.DisableAction);
            _c.RegisterAction(newActions.SetActiveAction);
            _c.RegisterAction(newActions.PlayAction);
            _c.RegisterAction(newActions.QueueAction);
            _c.RegisterAction(newActions.PlayRandomAction);
            _c.RegisterAction(newActions.QueueRandomAction);
            _c.RegisterAction(newActions.PlayNextAction);
            _c.RegisterAction(newActions.PlayPreviousAction);
            _c.RegisterAction(newActions.PlayCurrentAction);
            _c.RegisterAction(newActions.ResetPlayIndexAction);
            _collectionActions.Add(collectionName, newActions);
        }

        private void UpdateCollectionActionNames(string oldCollectionName, string newCollectionName)
        {
            if (string.IsNullOrEmpty(oldCollectionName) || string.IsNullOrEmpty(newCollectionName))
            {
                Log($"{nameof(AddNewCollectionActions)}: one or both input values are empty");
                return;
            }
            if (_collectionActions == null) return;
            CollectionActions actions;
            if (!_collectionActions.TryGetValue(oldCollectionName, out actions)) return;
            if (actions != null)
            {
                RemoveCollectionActions(oldCollectionName);
            }
            AddNewCollectionActions(newCollectionName);
        }

        private void RemoveCollectionActions(string collectionName)
        {
            CollectionActions actions;
            if (!_collectionActions.TryGetValue(collectionName, out actions)) return;
            _c.DeregisterAction(actions.EnableAction);
            _c.DeregisterAction(actions.DisableAction);
            _c.DeregisterAction(actions.SetActiveAction);
            _c.DeregisterAction(actions.PlayAction);
            _c.DeregisterAction(actions.QueueAction);
            _c.DeregisterAction(actions.PlayRandomAction);
            _c.DeregisterAction(actions.QueueRandomAction);
            _c.DeregisterAction(actions.PlayNextAction);
            _c.DeregisterAction(actions.PlayPreviousAction);
            _c.DeregisterAction(actions.PlayCurrentAction);
            _c.DeregisterAction(actions.ResetPlayIndexAction);
            _collectionActions.Remove(collectionName);
            actions.EnableAction = null;
            actions.DisableAction = null;
            actions.SetActiveAction = null;
            actions.PlayAction = null;
            actions.QueueAction = null;
            actions.PlayRandomAction = null;
            actions.QueueRandomAction = null;
            actions.PlayNextAction = null;
            actions.PlayPreviousAction = null;
            actions.PlayCurrentAction = null;
            actions.ResetPlayIndexAction = null;
        }

        public void Update()
        {
            if (_collections == null) return;
            foreach (var collection in _collections)
            {
                collection.Update();
            }
        }

        public static string GetEnableCollectionActionName(string collectionName)
        {
            return Tools.TrimAll($"{Storables.EnableCollectionAction}{collectionName}");
        }

        public static string GetDisableCollectionActionName(string collectionName)
        {
            return Tools.TrimAll($"{Storables.DisableCollectionAction}{collectionName}");
        }

        public static string GetSetActiveCollectionActionName(string collectionName)
        {
            return Tools.TrimAll($"{Storables.SetActiveCollectionAction}{collectionName}");
        }


        public static string GetPlayClipActionName(string collectionName)
        {
            return Tools.TrimAll($"{Storables.PlayClipFromCollectionAction}{collectionName}");
        }

        public static string GetQueueClipActionName(string collectionName)
        {
            return Tools.TrimAll($"{Storables.QueueClipFromCollectionAction}{collectionName}");
        }

        public static string GetPlayRandomClipActionName(string collectionName)
        {
            return Tools.TrimAll($"{Storables.PlayRandomClipFromCollectionAction}{collectionName}");
        }

        public static string GetQueueRandomClipActionName(string collectionName)
        {
            return Tools.TrimAll($"{Storables.QueueRandomClipFromCollectionAction}{collectionName}");
        }

        public static string GetPlayNextClipActionName(string collectionName)
        {
            return Tools.TrimAll($"{Storables.PlayNextClipFromCollectionAction}{collectionName}");
        }

        public static string GetPlayPreviousClipActionName(string collectionName)
        {
            return Tools.TrimAll($"{Storables.PlayPreviousClipFromCollectionAction}{collectionName}");
        }

        public static string GetPlayCurrentClipActionName(string collectionName)
        {
            return Tools.TrimAll($"{Storables.PlayCurrentClipFromCollectionAction}{collectionName}");
        }

        public static string GetResetPlayIndexActionName(string collectionName)
        {
            return Tools.TrimAll($"{Storables.ResetPlayIndexOfCollectionAction}{collectionName}");
        }

        #region Callbacks

        public void OnAtomRename(AudioMateController.ContainingAtomNameChangedEventArgs args)
        {
            var oldID = args.Before;
            var newID = args.After;
            foreach (var collection in _collections.Where(collection => collection.ReceiverAtomUID == oldID))
            {
                collection.OnAtomRename(oldID, newID);
            }
        }

        public void OnDestroy()
        {
            Log("Destroying Collection Manager");
            foreach (var collection in _collections)
            {
                if (collection != null) RemoveCollection(collection);
            }
            OnActiveCollectionSelected.RemoveAllListeners();
            OnActiveCollectionUpdated.RemoveAllListeners();
            OnActiveCollectionNameChanged.RemoveAllListeners();
            if (_autoPlayRoutine != null) StopCoroutine(_autoPlayRoutine);
            _autoPlayRoutine = null;
            if (_coroutine != null) StopCoroutine(_coroutine);
        }
        #endregion

        #region Serializer

        public void AddToJSON(JSONClass json, string id, string collectionName = null, bool excludeTriggerActions = false)
        {
            if (json == null) json = new JSONClass {{id, null}};
            if (json[id] == null) json.Add(id, null);
            json[id] = GetCollectionsJSON(collectionName, excludeTriggerActions);
        }

        private JSONClass GetCollectionsJSON(string collectionName = null, bool excludeTriggerActions = false)
        {
            var json = new JSONClass();
            try
            {
                var query = _collections.Where(collection => collection != null);
                if (collectionName != null) query = query.Where(collection => collection.Name == collectionName);
                foreach (var collection in query)
                {
                    json[collection.Name] = collection.GetJSON(excludeTriggerActions);
                }
            }
            catch (Exception e)
            {
                SuperController.LogError($"AudioMate.{nameof(CollectionManager)}.{nameof(GetCollectionsJSON)}: {e}");
            }
            return json;
        }

        private Coroutine _coroutine;
        public void RestoreFromJSON(JSONNode json, bool merge = false, bool createMissingAudioSources = false, string activeCollection = null)
        {
            Log("RestoreFromJSON in Collection Manager");
            if (json == null || !json.Childs.Any())
            {
                SuperController.LogError($"AudioMate.{nameof(CollectionManager)}.{nameof(RestoreFromJSON)}: Can not parse JSON data because it's empty.");
                return;
            }

            if (isParsingCollections) return;

            try
            {
                /* if (_coroutine == null)
                { */
                    _coroutine = StartCoroutine(RestoreCollections(json.Childs.ToList(), merge, createMissingAudioSources, activeCollection));
                /*}
                else
                {
                    StopCoroutine(_coroutine);
                    _coroutine = null;
                    _coroutine = StartCoroutine(RestoreCollections(json.Childs.ToList(), merge, createMissingAudioSources, activeCollection));
                } */
            }
            catch (Exception e)
            {
                SuperController.LogError($"AudioMate.{nameof(CollectionManager)}.{nameof(RestoreFromJSON)}: {e}");
            }
        }

        private IEnumerator RestoreCollections(List<JSONNode> collectionList, bool merge, bool createMissingAudioSources, string activeCollection = null)
        {
            if (_collections == null) yield break;
            isParsingCollections = true;
            DisableAll();
            if (merge == false)
            {
                for (var index = 0; index < _collections.Count; index++)
                {
                    var collection = _collections[index];
                    if (collection == null) continue;
                    RemoveCollection(collection);
                }
                _collections.Clear();
            }

            foreach (var collectionJSON in collectionList)
            {
                yield return null;
                var mergeTarget = _collections.Find(x => x.Name == collectionJSON["name"].Value);
                var collection = mergeTarget ?? new Collection(_c, collectionJSON["name"].Value);
                var successfulImport = collection.RestoreFromJSON(collectionJSON.AsObject, merge, true);
                if (!successfulImport) continue;
                if (collection.ReceiverAtomUID != null && createMissingAudioSources)
                {
                    var audioSource = SuperController.singleton.GetAtomByUid(collection.ReceiverAtomUID);
                    if ((Object) audioSource == (Object) null)
                    {
                        StartCoroutine(_c.CreateAtom("AudioSource", collection.ReceiverAtomUID, newAtom =>
                        {
                            var hip = _c.containingAtom.GetStorableByID("hip");
                            _c.ParentLinkAtPosition(newAtom, _c.containingAtom, hip, "hip");
                            audioSource = newAtom;
                            var audioSourceControl = audioSource.GetStorableByID("AudioSource") as AudioSourceControl;
                            if ((Object) audioSourceControl == (Object) null) return;
                            audioSourceControl.spatialize = true;
                            audioSourceControl.volume = 1.0f;
                            audioSourceControl.pitch = 1.0f;
                        }));
                    }
                }

                yield return null;
                collection.Init();
                AddCollection(collection);
            }

            yield return null;
            EnableAll();
            SelectActiveCollection(string.IsNullOrEmpty(activeCollection) ? _collections?.First()?.Name : activeCollection);
            _c.needsProvisionalClipProcessing = true;
            isParsingCollections = false;
        }

        #endregion

        private void Log(string message)
        {
            if ((UnityEngine.Object) _c == (UnityEngine.Object) null) return;
            _c.Log($"CollectionManager: {message}");
        }
    }
}
