﻿using System.Collections.Generic;
using System.Globalization;
using System.Linq;
using Leap.Unity;
using SimpleJSON;
using UnityEngine;
using UnityEngine.Experimental.PlayerLoop;

namespace AudioMate
{

    public static class TriggerCategories
    {
        public const string Collision = "collisionTrigger";
        public const string Volume = "volumeTrigger";
    }

    public static class TriggerActionTypes
    {
        public const string Start = "startActions";
        public const string Transition = "transitionActions";
        public const string End = "endActions";
        public const string All = "allActions";

        public static List<string> AsList
        {
            get
            {
                var result = new List<string>
                {
                    Start,
                    Transition,
                    End
                };

                return result;
            }
        }
    }

    public class CollectionTrigger : Trigger
    {
        private AudioMateController _c;
        private Atom _containingAtom;
        private Transform _transform;
        public Collection Collection;
        private float _clipLoudness;
        private float _smoothedClipLoudness;
        private float _volumeTriggerMultiplier = 5f;
        private float _volumeTriggerQuickness = 15f;

        private readonly string _category;

        public CollectionTrigger(AudioMateController controller, string category, string sourceUID = null)
        {
            if (category == null)
            {
                SuperController.LogError("Can not init CollectionTrigger because the category is null.");
                return;
            }
            _category = category;
            if ((Object) controller == (Object) null)
            {
                SuperController.LogError("Can not init CollectionTrigger because AudioMateController Instance is null.");
                return;
            }
            if ((Object) controller.containingAtom == (Object) null)
            {
                SuperController.LogError("Can not init CollectionTrigger because the containing Atom of the AudioMateController instance is null.");
                return;
            }

            Init(controller);
        }

        private void Init(AudioMateController controller)
        {
            _c = controller;
            _containingAtom = _c.containingAtom;
            _transform = _c.UITransform;
            triggerActionsParent = _transform;
            CreateFloatJSON();
            InitTriggerUI();
            InitTriggerActionsUI();
        }

        public override JSONClass GetJSON(string subScenePrefix)
        {
            var jc = base.GetJSON(subScenePrefix);
            foreach (var tAType in TriggerActionTypes.AsList)
            {
                if (!jc.HasKey(tAType)) continue;
                var asArray = jc[tAType].AsArray;
                for (var i = 0; i < asArray.Count; i++)
                {
                    var triggerAction = asArray[i].AsObject;
                    Collection.InsertVariables(triggerAction);
                    asArray[i] = triggerAction;
                }
                jc[tAType] = asArray;
            }

            return jc;
        }

        public override void RestoreFromJSON(JSONClass jc, string subScenePrefix, bool isMerge)
        {
            foreach (var tAType in TriggerActionTypes.AsList)
            {
                if (!jc.HasKey(tAType)) continue;
                var asArray = jc[tAType].AsArray;
                for (var i = 0; i < asArray.Count; i++)
                {
                    var triggerAction = asArray[i].AsObject;
                    Collection.ParseVariables(triggerAction);
                    asArray[i] = triggerAction;
                }
                jc[tAType] = asArray;
            }
            base.RestoreFromJSON(jc, subScenePrefix, isMerge);
        }

        private bool _prevActive;
        public void SetActiveState(bool isActive)
        {
            if (_category != TriggerCategories.Collision) return;
            _prevActive = _active;
            _active = isActive;
            if (_prevActive == _active) return;
            if (_active)
            {
                foreach (var triggerActionDiscrete in discreteActionsStart)
                {
                    triggerActionDiscrete.Trigger();
                }
            }
            else
            {
                foreach (var triggerActionDiscrete in discreteActionsEnd)
                {
                    triggerActionDiscrete.Trigger();
                }
            }

        }

        public void CreatePlayCollectionTriggerAction(string actionName, string actionType = TriggerActionTypes.Start, int index = -1)
        {
            if (Collection == null)
            {
                SuperController.LogError("CollectionTrigger.CreatePlayCollectionTriggerAction: Collection Trigger has no valid Collection.");
                return;
            }
            var receiverAtom = _containingAtom;
            if ((Object) receiverAtom == (Object) null) return;

            var triggerAction = actionType == TriggerActionTypes.Start
                ? CreateDiscreteActionStartInternal(index)
                : CreateDiscreteActionEndInternal(index);
            if (triggerAction == null) return;

            triggerAction.handler = this;
            triggerAction.name = actionName;
            triggerAction.receiverAtom = receiverAtom;
            triggerAction.receiver = _c;
            triggerAction.receiverTargetName = CollectionManager.GetPlayClipActionName(Collection.Name);
        }

        public List<TriggerActionTransition> VolumeTriggeredMorphActions()
        {
            switch (_category)
            {
                case TriggerCategories.Collision:
                    return null;
                case TriggerCategories.Volume:
                    return transitionActions;
                default:
                    return null;
            }
        }

        public List<TriggerActionDiscrete> CollisionTriggerActionsStart()
        {
            switch (_category)
            {
                case TriggerCategories.Collision:
                    return discreteActionsStart;
                case TriggerCategories.Volume:
                    return null;
                default:
                    return null;
            }
        }

        public List<TriggerActionDiscrete> CollisionTriggerActionsEnd()
        {
            switch (_category)
            {
                case TriggerCategories.Collision:
                    return discreteActionsEnd;
                case TriggerCategories.Volume:
                    return null;
                default:
                    return null;
            }
        }

        public bool HasTriggerActions()
        {
            return discreteActionsStart.Count > 0
                   || transitionActions.Count > 0
                   || discreteActionsEnd.Count > 0;
        }

        public TriggerActionTransition CreateVolumeTriggeredMorphAction(JSONClass json)
        {
            if (json == null || !json.Childs.Any() || json["name"].Value == null || json["receiverTarget"].Value == null) return null;
            float startVal;
            if (!float.TryParse(json["startVal"].Value, out startVal)) startVal = 0f;
            float endVal;
            if (!float.TryParse(json["endVal"].Value, out endVal)) endVal = 1f;
            bool isActive;
            if (!bool.TryParse(json["active"].Value, out isActive)) isActive = true;
            return CreateVolumeTriggeredMorphAction(json["name"].Value, json["receiverTarget"].Value, -1, startVal, endVal,
                isActive);
        }

        public TriggerActionTransition CreateVolumeTriggeredMorphAction(string actionName, string morph, int index = -1, float startVal = 0f, float endVal = 1f, bool isActive = true)
        {
            if (Collection == null)
            {
                SuperController.LogError("CollectionTrigger.CreateVolumeTriggeredMorphAction: Collection Trigger has no valid Collection.");
                return null;
            }
            var audioSource = Collection.ReceiverAudioSource;
            if ((Object) audioSource == (Object) null)
            {
                SuperController.LogError("CollectionTrigger.CreateVolumeTriggeredMorphAction: Receiver Audio Source is null.");
                return null;
            }

            var receiverAtom = _containingAtom;
            if ((Object) receiverAtom == (Object) null) return null;
            var receiverGeometry = receiverAtom.GetStorableByID("geometry");
            if ((Object) receiverGeometry == (Object) null) return null;

            var transitionAction = CreateTransitionActionInternal(index);
            if (transitionAction == null) return null;

            transitionAction.handler = this;
            transitionAction.name = actionName;
            transitionAction.receiverAtom = receiverAtom;
            transitionAction.receiver = receiverGeometry;
            transitionAction.receiverTargetName = morph;
            transitionAction.startVal = startVal;
            transitionAction.endVal = endVal;
            transitionAction.active = isActive;
            transitionAction.TriggerInterp(transitionInterpValue);

            return transitionAction;
        }

        public TriggerActionTransition GetTransitionActionByReceiver(Atom receiverAtom, JSONStorable receiver = null, string receiverTargetName = null)
        {
            if ((Object) receiverAtom == (Object) null) return null;
            for (var index = 0; index < transitionActions.Count; index++)
            {
                var triggerActionTransition = transitionActions[index];
                if (triggerActionTransition == null) continue;

                if ((Object) triggerActionTransition.receiverAtom != (Object) null && triggerActionTransition.receiverAtom != receiverAtom) continue;

                if ((Object) receiver != (Object) null && triggerActionTransition.receiver != receiver) continue;

                if (receiverTargetName != null &&
                    triggerActionTransition.receiverTargetName != receiverTargetName) continue;

                return triggerActionTransition;
            }

            return null;
        }

        public void SetVolumeTriggerMultiplier(float value)
        {
            _volumeTriggerMultiplier = value;
        }

        public void SetVolumeTriggerQuickness(float value)
        {
            _volumeTriggerQuickness = value;
        }

        public override void Update()
        {
            if (_category == TriggerCategories.Collision) return;
            switch (_category)
            {
                /*case TriggerCategories.Collision:
                {
                    foreach (TriggerActionDiscrete triggerActionDiscrete in discreteActionsStart)
                        triggerActionDiscrete.Update();
                    foreach (TriggerActionDiscrete triggerActionDiscrete in discreteActionsEnd)
                        triggerActionDiscrete.Update();
                    break;
                } */
                case TriggerCategories.Volume:
                {
                    if (!Collection.IsPlaying()) break;
                    _clipLoudness = Collection.ReceiverAudioSource.clipLoudness;
                    _smoothedClipLoudness = Mathf.Lerp(floatJSON.val, Mathf.Clamp01(_clipLoudness * _volumeTriggerMultiplier), Time.deltaTime * _volumeTriggerQuickness);
                    floatJSON.val = _smoothedClipLoudness;
                    for (var index = 0; index < transitionActions.Count; index++)
                    {
                        var triggerActionTransition = transitionActions[index];
                        triggerActionTransition?.TriggerInterp(transitionInterpValue, true);
                    }

                    break;
                }
            }
        }

        protected JSONStorableFloat floatJSON;

        protected void SyncFloat(float f)
        {
            if (_category == TriggerCategories.Volume) {

                transitionInterpValue = floatJSON.val;
            }
        }

        protected virtual void CreateFloatJSON()
        {
            floatJSON = new JSONStorableFloat("value", 0.0f, SyncFloat, 0.0f, 1f);
            floatJSON.storeType = JSONStorableParam.StoreType.Physical;
            _c.RegisterFloat(this.floatJSON);
        }
    }
}
