﻿using System.Collections.Generic;
using System.Globalization;
using UnityEngine;
using UnityEngine.Serialization;

namespace AudioMate
{
    public class ScreenClips : Screen
    {
        public new static readonly ScreenDefinition Definition = new ScreenDefinition(
            "Clips",
            "Clip Library",
            typeof(ScreenClips));

        private HorizontalButtonGroup _btnButtonGroupAddClips;
        private UIClipLibrary _clipLibrary;
        private UIDynamicButton _btnShowMode;

        public const string FilterAll = "All Clips";
        public const string FilterUnassigned = "Unassigned Clips";
        public const string FilterAssigned = "Assigned Clips";
        public List<string> filterModes = new List<string> {FilterAll, FilterUnassigned, FilterAssigned};
        private string _activeFilter = FilterAll;

        private List<UIDynamicButton> _bulkSelectButtons;

        public override void Init(AudioMateController controller)
        {
            base.Init(controller);
            UI.CreateSpacer().height = 15f;
            InitDisplayAndSelectionUI();
            CreateClipLibraryUI();
            InitFileImportUI();
            controller.OnBulkModeOff.AddListener(OnBulkSelectOff);
        }

        private void InitFileImportUI()
        {
            _bulkSelectButtons = new List<UIDynamicButton>();
            _btnButtonGroupAddClips = UI.CreateHorizontalGroup();
            _btnButtonGroupAddClips.CreateLabel("Select", Styles.Button, false, 28, false, TextAnchor.MiddleLeft, 80f);
            var add10Btn = _btnButtonGroupAddClips.CreateButton(C.FirstMultiselectButtonAmountJSON.val.ToString("F0", CultureInfo.InvariantCulture));
            add10Btn.button.onClick.AddListener(() => OnBulkSelect((int) C.FirstMultiselectButtonAmountJSON.val, add10Btn));
            _bulkSelectButtons.Add(add10Btn);
            var add20Btn = _btnButtonGroupAddClips.CreateButton(C.SecondMultiselectButtonAmountJSON.val.ToString("F0", CultureInfo.InvariantCulture));
            _bulkSelectButtons.Add(add20Btn);
            add20Btn.button.onClick.AddListener(() => OnBulkSelect((int) C.SecondMultiselectButtonAmountJSON.val, add20Btn));
            _btnButtonGroupAddClips.CreateButton("All", Styles.Button, () =>
            {
                if (_activeFilter == FilterAll)
                {
                    C.collections.AddAllClipsToActiveCollection();
                    return;
                }
                C.collections.AddVisibleClipsToActiveCollection();
            });
            var btnGroupImport = UI.CreateHorizontalGroup();
            btnGroupImport.CreateLabel("Import", Styles.Button, false, 28, false, TextAnchor.MiddleLeft, 0f);
            btnGroupImport.CreateButton($"{Symbols.Add} File", Styles.Button, () => C.audioFileManager.OpenImportFileDialog(C.AddImportedFilesToActiveCollectionJSON.val));
            btnGroupImport.CreateButton($"{Symbols.Add} Folder", Styles.Button, () => C.audioFileManager.OpenImportFolderDialog(C.AddImportedFilesToActiveCollectionJSON.val));

            var btnGroupRemove = UI.CreateHorizontalGroup();
            btnGroupRemove.CreateLabel("Delete", Styles.Button, false, 28, false, TextAnchor.MiddleLeft, 0f);
            btnGroupRemove.CreateButton($"{Symbols.Remove} Unassigned Clips", Styles.Danger, () =>
            {
                C.ui.ShowYesNoModal(
                    "This operation removes all clips from the scene that are not assigned to a collection. Are you sure?",
                    C.clipLibrary.RemoveUnusedClips,
                    null, "Warning"
                    );
            });

            btnGroupRemove.CreateButton($"{Symbols.Remove} All Clips", Styles.Danger, () =>
            {
                C.ui.ShowYesNoModal(
                    "This operation removes all clips from the scene and empties all of your collections. Are you sure?",
                    C.clipLibrary.RemoveAllClips,
                    null, "Warning"
                );
            });

        }

        private void CreateClipLibraryUI()
        {
            _clipLibrary = UIClipLibrary.AddTo(UI.transform);
            _clipLibrary.Init(C);
            _clipLibrary.Bind();
            _clipLibrary.Refresh();
        }

        private void InitDisplayAndSelectionUI()
        {
            var btnGroupFilter = UI.CreateHorizontalGroup();
            btnGroupFilter.CreateLabel("Show");
            _btnShowMode = btnGroupFilter.CreateButton($"{Symbols.Selection} {_activeFilter}", Styles.Button, () =>
            {
                var nextIndex = filterModes.IndexOf(_activeFilter)+1;
                if (nextIndex > filterModes.Count - 1) nextIndex = 0;
                _activeFilter = filterModes[nextIndex];
                _clipLibrary.SetFilter(_activeFilter);
                _btnShowMode.buttonText.text = $"{Symbols.Selection} {_activeFilter}";
            });
        }

        private void OnBulkSelect(int amount, UIDynamicButton button)
        {
            if (C.isBulkSelectOn)
            {
                C.isBulkSelectOn = false;
                if (_bulkSelectButtons.Count == 0) return;
                if (amount == C.bulkSelectAmount) return;
                foreach (var bulkSelectButton in _bulkSelectButtons)
                {
                    bulkSelectButton.buttonColor = Styles.ButtonBg;
                    bulkSelectButton.textColor = Styles.ButtonText;
                }
            }
            C.isBulkSelectOn = true;
            C.bulkSelectAmount = amount;
            button.buttonColor = Styles.ButtonSelectedBg;
            button.textColor = Styles.ButtonSelectedText;
        }

        private void OnBulkSelectOff()
        {
            foreach (var bulkSelectButton in _bulkSelectButtons)
            {
                bulkSelectButton.buttonColor = Styles.ButtonBg;
                bulkSelectButton.textColor = Styles.ButtonText;
            }

        }

        public override void OnDestroy()
        {
            if (C != null)
            {
                C.OnBulkModeOff.RemoveListener(OnBulkSelectOff);
            }
            base.OnDestroy();
        }
    }
}
