﻿using System;
using System.Collections.Generic;
using System.Linq;
using SimpleJSON;
using UnityEngine;
using UnityEngine.UI;
using Object = UnityEngine.Object;

namespace AudioMate
{
    public class ScreenCollection : Screen
    {
        public new static readonly ScreenDefinition Definition = new ScreenDefinition(
            "Collection",
            "Collection",
            typeof(ScreenCollection));

        private JSONStorableStringChooser _popupLayerJSON;

        private JSONStorableBool _toggleEnabledJSON;
        private JSONStorableBool _toggleOnlyIfClearJSON;
        private JSONStorableBool _toggleOnlyIfThisCollectionIsNotPlayingJSON;
        private JSONStorableBool _toggleIgnoreLayersJSON;
        private JSONStorableBool _toggleAlwaysQueueJSON;
        private JSONStorableBool _toggleShuffleJSON;
        private JSONStorableBool _toggleAutoPlayJSON;
        private JSONStorableBool _toggleContinuousPlayJSON;
        private JSONStorableBool _toggleLoopJSON;

        private JSONStorableFloat _sliderVolumeJSON;
        private JSONStorableFloat _sliderPitchJSON;
        private JSONStorableFloat _sliderPlayChanceJSON;
        private JSONStorableFloat _sliderVolumeTriggerQuicknessJSON;
        private JSONStorableFloat _sliderVolumeTriggerMultiplierJSON;
        private JSONStorableFloat _sliderMorphTestJSON;

        private Text _textHeading;

        private bool _needsRefresh;

        private readonly List<HorizontalButtonGroup> _triggerActionsUI = new List<HorizontalButtonGroup>();
        private readonly List<HorizontalButtonGroup> _volumeTriggerUI = new List<HorizontalButtonGroup>();

        private GameObject _collisionTriggerActionsUI;
        private GameObject _volumeTriggerActionsUI;

        public override void Init(AudioMateController controller)
        {
            base.Init(controller);
            CreateMainUI();
        }

        private void InitSettings(Collection activeCollection)
        {
            if (activeCollection == null) return;
            _textHeading.text = $"Settings for '{activeCollection.Name}'";
            _toggleEnabledJSON.val = activeCollection.Enabled;
            _popupLayerJSON.val = activeCollection.Layer.ToString();
            _toggleIgnoreLayersJSON.val = activeCollection.IgnoreLayers;
            _toggleOnlyIfClearJSON.val = activeCollection.OnlyIfClear;
            _toggleOnlyIfThisCollectionIsNotPlayingJSON.val = activeCollection.OnlyIfThisCollectionIsNotPlaying;
            _toggleAlwaysQueueJSON.val = activeCollection.AlwaysQueue;
            _toggleShuffleJSON.val = activeCollection.ShuffleMode;
            _toggleLoopJSON.val = activeCollection.Loop;
            _toggleAutoPlayJSON.val = activeCollection.AutoPlay;
            _toggleContinuousPlayJSON.val = activeCollection.ContinuousPlay;
            if (C.VolumeControlEnabledJSON.val) _sliderVolumeJSON.val = activeCollection.Volume;
            if (C.PitchControlEnabledJSON.val) _sliderPitchJSON.val = activeCollection.Pitch;
            _sliderPlayChanceJSON.val = activeCollection.PlayChance;
            if (_sliderVolumeTriggerMultiplierJSON != null) _sliderVolumeTriggerMultiplierJSON.val = activeCollection.VolumeTriggerMultiplier;
            if (_sliderVolumeTriggerQuicknessJSON != null) _sliderVolumeTriggerQuicknessJSON.val = activeCollection.VolumeTriggerQuickness;
        }

        private void CreateMainUI()
        {
            _textHeading = UI.CreateHeader("Settings for 'Untitled'", 2, 20f);
            if (C.DisplayHints) {
                UI.CreateText("These are the settings for the currently active (selected) collection. Use the drop-down and the buttons on the right side to select, create, clear or delete a collection.", FontSizes.Small);
                UI.CreateText("To change the name of this collection, click on the big white name on the right side and enter a new one.", FontSizes.Small);
            }
            _toggleEnabledJSON = new JSONStorableBool("Enabled", true, (bool val) =>
            {
                ActiveCollection.Enabled = val;
            });
            UI.CreateToggle(_toggleEnabledJSON);
            UI.CreateSpacer(3f);
            if (C.VolumeControlEnabledJSON.val)
            {
                _sliderVolumeJSON = new JSONStorableFloat("Volume", 1f,
                    (float val) => { ActiveCollection.Volume = val; }, 0f, 1f);
                UI.CreateSlider(_sliderVolumeJSON, 1F, true);
                UI.CreateSpacer(3f);
            }
            if (C.PitchControlEnabledJSON.val)
            {
                _sliderPitchJSON = new JSONStorableFloat("Pitch", 1f,
                    (float val) => { ActiveCollection.Pitch = val; }, 0f, 2f);
                UI.CreateSlider(_sliderPitchJSON, 1f, true);
                UI.CreateSpacer(3f);
            }
            _sliderPlayChanceJSON = new JSONStorableFloat("Play Chance", 1f, (float val) =>
            {
                ActiveCollection.PlayChance = val;
            }, 0f, 1f);
            UI.CreateSlider(_sliderPlayChanceJSON, 1f, true);
            UI.CreateSpacer(3f);

            _popupLayerJSON = new JSONStorableStringChooser("Layer", GetAvailableLayerOptions(), ActiveCollection.Layer.ToString(), "Layer", val => ActiveCollection.Layer = int.Parse(val));
            UI.CreatePopup(_popupLayerJSON, false);

            _toggleIgnoreLayersJSON = new JSONStorableBool("Ignore Layers", false, (bool val) =>
            {
                    ActiveCollection.IgnoreLayers = val;
            });
            UI.CreateToggle(_toggleIgnoreLayersJSON);
            if (C.DisplayHints)
            {
                UI.CreateText(
                    "Collection will not play if a collection on a higher layer is playing on the same audio receiver. You can disable this behaviour by checking 'Ignore Layers'.");
            }

            _toggleOnlyIfClearJSON = new JSONStorableBool("Play only if audio receiver is clear", false, (bool val) =>
            {
                ActiveCollection.OnlyIfClear = val;
                if (val)
                {
                    _toggleOnlyIfThisCollectionIsNotPlayingJSON.val = false;
                }
            });
            UI.CreateToggle(_toggleOnlyIfClearJSON);

            _toggleOnlyIfThisCollectionIsNotPlayingJSON = new JSONStorableBool("Play only if this collection isn't playing", false, (bool val) =>
            {
                ActiveCollection.OnlyIfThisCollectionIsNotPlaying = val;
                if (val)
                {
                    _toggleOnlyIfClearJSON.val = false;
                }
            });
            UI.CreateToggle(_toggleOnlyIfThisCollectionIsNotPlayingJSON);

            _toggleAlwaysQueueJSON = new JSONStorableBool("Always queue", false, (bool val) =>
            {
                ActiveCollection.AlwaysQueue = val;
            });
            UI.CreateToggle(_toggleAlwaysQueueJSON);

            _toggleShuffleJSON = new JSONStorableBool("Shuffle mode (Random order playback)", true, (bool val) =>
            {
                ActiveCollection.ShuffleMode = val;
            });
            UI.CreateToggle(_toggleShuffleJSON);

            _toggleLoopJSON = new JSONStorableBool("Loop (After each clip has been played)", true, (bool val) =>
            {
                ActiveCollection.Loop = val;
            });
            UI.CreateToggle(_toggleLoopJSON);

            _toggleAutoPlayJSON = new JSONStorableBool("Autoplay (On scene load)", false, (bool val) =>
            {
                ActiveCollection.HasAutoPlayBeenTriggered = false;
                ActiveCollection.AutoPlay = val;
            });
            UI.CreateToggle(_toggleAutoPlayJSON);

            _toggleContinuousPlayJSON = new JSONStorableBool("Continuous play (After first play)", false, (bool val) =>
            {
                ActiveCollection.ContinuousPlay = val;
            });
            UI.CreateToggle(_toggleContinuousPlayJSON);

            var activeCollection = ActiveCollection;
            InitSettings(activeCollection);
            CreateTriggerActionsUI();
            CreateVolumeTriggeredMorphActionsUI();
        }

        public List<string> GetAvailableLayerOptions()
        {
            return new List<string> {"0", "1", "2", "3", "4", "5", "6", "7", "8", "9", "10", "11", "12", "13", "14", "15", "16", "17", "18", "19", "20"};
        }

        #region Collision Trigger Actions
        private void CreateTriggerActionsUI()
        {
            UI.CreateHeader("Physical Trigger Actions");
            if (C.DisplayHints) {
                UI.CreateText("Add and remove bindings to physical trigger actions here. If a binding is active, it will play this collection when the selected trigger is fired.", FontSizes.Small);
                UI.CreateText("'Start Actions' will run as soon as the trigger is fired, 'End Actions' will run when the trigger has stopped.", FontSizes.Small);
            }
            UI.CreatePopup(C.TriggerColliderChooserJSON, true, 250f);
            C.TriggerColliderChooserJSON.setCallbackFunction = val =>
            {
                C.TriggerColliderChooserJSON.choices = C.ColliderChoices;
                _needsRefresh = true;
            };

            _collisionTriggerActionsUI = UI.CreateLayoutContainer();
            var layoutGroup = _collisionTriggerActionsUI.GetComponent<VerticalLayoutGroup>();
            layoutGroup.spacing = 10f;
            _needsRefresh = true;
        }

        private HorizontalButtonGroup _addStartActionButton;
        private void CreateAddStartActionButton()
        {
            if ((Object) _addStartActionButton != (Object) null) return;
            _addStartActionButton = UI.CreateHorizontalGroup(_collisionTriggerActionsUI);
            _addStartActionButton.Padding = new RectOffset(0, 0, 0, 0);
            var button = _addStartActionButton.CreateButton($"{Symbols.Add} Play On Trigger Start", Styles.Button, () =>
            {
                OnAddTriggerButtonClicked();
                _needsRefresh = true;
            });
            button.height = 70f;
        }

        private HorizontalButtonGroup _addEndActionButton;
        private void CreateAddEndActionButton()
        {
            if ((Object) _addEndActionButton != (Object) null) return;
            _addEndActionButton = UI.CreateHorizontalGroup(_collisionTriggerActionsUI);
            _addEndActionButton.Padding =  new RectOffset(0, 0, 0, 0);
            var button = _addEndActionButton.CreateButton($"{Symbols.Add} Play On Trigger End", Styles.Button, () =>
            {
                OnAddTriggerButtonClicked(TriggerActionTypes.End);
                _needsRefresh = true;
            });
            button.height = 70f;
        }

        private void CreateTriggerActionUI(CollectionTrigger collectionTrigger, TriggerAction triggerAction, string triggerActionType)
        {
            var layoutGroup = UI.CreateHorizontalGroup(_collisionTriggerActionsUI);
            layoutGroup.Padding = new RectOffset(20, 10, 10, 10);
            layoutGroup.SetBg(Color.white);
            layoutGroup.CreateLabel(triggerActionType == TriggerActionTypes.Start
                ? "Plays On Trigger Start"
                : "Plays On Trigger End", Styles.Button, false, 26, true, TextAnchor.MiddleLeft);

            layoutGroup.CreateButton(Symbols.Remove, Styles.Danger, () =>
            {
                _needsRefresh = true;
                collectionTrigger.RemoveTriggerAction(triggerAction);
            }, false, 50f, 0);

            _triggerActionsUI.Add(layoutGroup);
        }

        public void OnAddTriggerButtonClicked(string triggerType = TriggerActionTypes.Start)
        {
            if ((Object) Collections == (Object) null)
            {
                Log("Can not add Trigger Action because Collections is null.");
                return;
            }

            if (ActiveCollection == null)
            {
                Log("Can not add Trigger Action because Active Collection is null.");
                return;
            }
            var triggerUID = C.TriggerColliderChooserJSON.val;
            ActiveCollection.GetCollisionTrigger(triggerUID)
                .CreatePlayCollectionTriggerAction($"{Storables.PlayClipFromCollectionAction}{ActiveCollection.Name}",
                    triggerType);
        }

        private void RefreshCollisionTriggerActionsUI()
        {
            var selectedTrigger = C.TriggerColliderChooserJSON.val;
            if (selectedTrigger == "None" || selectedTrigger == null) return;
            C.TriggerColliderChooserJSON.choices = C.ColliderChoices;
            C.TriggerColliderChooserJSON.displayChoices = C.ColliderChoices;
            var collisionTrigger = ActiveCollection.GetCollisionTrigger(selectedTrigger);
            var triggerActionsStart = collisionTrigger.CollisionTriggerActionsStart();
            var triggerActionsEnd = collisionTrigger.CollisionTriggerActionsEnd();
            ResetCollisionTriggerActionsUI();
            if (triggerActionsStart?.Count > 0)
            {
                foreach (var triggerActionDiscrete in triggerActionsStart)
                {
                    CreateTriggerActionUI(collisionTrigger, triggerActionDiscrete, TriggerActionTypes.Start);
                }
            }
            else
            {
                CreateAddStartActionButton();
            }

            if (triggerActionsEnd?.Count > 0)
            {
                foreach (var triggerActionDiscrete in triggerActionsEnd)
                {
                    CreateTriggerActionUI(collisionTrigger, triggerActionDiscrete, TriggerActionTypes.End);
                }
            }
            else
            {
                CreateAddEndActionButton();
            }
        }

        private void ResetCollisionTriggerActionsUI()
        {
            foreach (var uiElem in _triggerActionsUI)
            {
                uiElem.Destroy();
            }
            _triggerActionsUI.Clear();
            if ((Object) _addStartActionButton != (Object) null) _addStartActionButton.Destroy();
            if ((Object) _addEndActionButton != (Object) null) _addEndActionButton.Destroy();
            _addStartActionButton = null;
            _addEndActionButton = null;
        }

        #endregion

        #region Volume Triggered Morphs

        private JSONStorableFloat _selectedMorph;
        private void CreateVolumeTriggeredMorphActionsUI()
        {
            UI.CreateHeader("Volume Triggered Morphs");
            if (C.DisplayHints) {
                UI.CreateText("Add and remove bindings to volume triggered morphs here. A volume triggered morph will set a morph value according to the current playback volume of a collection.", FontSizes.Small);
            }

            var audioSource = ActiveCollection?.ReceiverAudioSource;
            if ((Object) audioSource == (Object) null) return;

            _sliderVolumeTriggerMultiplierJSON = new JSONStorableFloat("Multiplier",ActiveCollection.VolumeTriggerMultiplier,
                (float val) => ActiveCollection.VolumeTriggerMultiplier = val, 0f, 100f, false);
            UI.CreateSlider(_sliderVolumeTriggerMultiplierJSON, 1F, true);

            _sliderVolumeTriggerQuicknessJSON = new JSONStorableFloat("Quickness", ActiveCollection.VolumeTriggerQuickness,
                (float val) => ActiveCollection.VolumeTriggerQuickness = val, 0f, 50f, false);
            UI.CreateSlider(_sliderVolumeTriggerQuicknessJSON, 1F, true);

            UI.CreateSpacer(3f);

            UI.CreatePopup(C.MorphChooserJSON, true, 250f);
            C.MorphChooserJSON.setCallbackFunction = val =>
            {
                if (C == null || C.containingAtom == null) return;
                C.MorphChooserJSON.choices = C.GetMorphLabelList();
                _selectedMorph = C.containingAtom.GetStorableByID("geometry").GetFloatJSONParam(val);
                if (_selectedMorph != null)
                {
                    _sliderMorphTestJSON.min = _selectedMorph.min;
                    _sliderMorphTestJSON.max = _selectedMorph.max;
                    _sliderMorphTestJSON.defaultVal = _selectedMorph.defaultVal;
                    _sliderMorphTestJSON.valNoCallback = _selectedMorph.val;

                }
                _needsRefresh = true;
            };

            _sliderMorphTestJSON = new JSONStorableFloat("Test", 0f,
                (float val) =>
                {
                    if (_selectedMorph != null) _selectedMorph.val = val;
                }, 0f, 1f, false);
            UI.CreateSlider(_sliderMorphTestJSON, 1F, true);

            _volumeTriggerActionsUI = UI.CreateLayoutContainer();
            var layout = _volumeTriggerActionsUI.GetComponent<VerticalLayoutGroup>();
            layout.spacing = 10f;
            UI.CreateSpacer(200f);
            _needsRefresh = true;
        }

        private void CreateVolumeTriggeredMorphActionUI(TriggerActionTransition triggerAction)
        {
            var layoutGroup = UI.CreateHorizontalGroup(_volumeTriggerActionsUI);
            layoutGroup.SetBg(Color.white);
            layoutGroup.Padding = new RectOffset(20, 10, 10, 10);

            layoutGroup.CreateLabel(triggerAction.receiverTargetName, Styles.Button, false, 26, true, TextAnchor.MiddleLeft);
            layoutGroup.CreateButton(Symbols.Edit, Styles.Button, triggerAction.OpenDetailPanel, false, 50f, 0);
            layoutGroup.CreateButton(Symbols.Remove, Styles.Danger, () =>
            {
                // Reset morph value to default before removing the trigger action
                var morph = C.GetMorph(triggerAction.receiverTargetName);
                if ((Object) triggerAction.receiver != (Object) null) triggerAction.receiver.SetFloatParamValue(triggerAction.receiverTargetName, morph.defaultVal);
                ActiveCollection.GetVolumeTrigger().RemoveTriggerAction(triggerAction);
                _volumeTriggerUI.Remove(layoutGroup);
                if ((Object) layoutGroup != (Object) null) layoutGroup.Destroy();
                layoutGroup = null;
                _needsRefresh = true;
            }, false, 50f, 0);
            _volumeTriggerUI.Add(layoutGroup);
        }

        private HorizontalButtonGroup _addVolumeTriggeredMorphActionButton;
        private bool _hasAskedToAdjustSettings;
        private void CreateAddVolumeTriggeredMorphActionButton()
        {
            if ((Object) _addVolumeTriggeredMorphActionButton != (Object) null) return;
            _addVolumeTriggeredMorphActionButton = UI.CreateHorizontalGroup(_volumeTriggerActionsUI);
            _addVolumeTriggeredMorphActionButton.Padding = new RectOffset(0, 0, 0, 0);
            var button = _addVolumeTriggeredMorphActionButton.CreateButton($"{Symbols.Add} {C.MorphChooserJSON.val}", Styles.Button, () =>
            {
                if (!_hasAskedToAdjustSettings && C.IsInterferingAutoBehaviourEnabled())
                {
                    C.ui.ShowYesNoModal("It is recommended to switch off automatic jaw and mouth movements when using volume triggered morphs. \n \n" +
                                          "Should I disable the settings 'Auto Jaw Mouth Morphs' and 'Drive Jaw Angle From Head Audio' of this Person Atom for you?",
                        () =>
                        {
                            C.DisableInterferingPersonAtomSettings();
                            OnAddVolumeTriggeredMorphButtonClicked();
                        }, OnAddVolumeTriggeredMorphButtonClicked, "Notice");
                    _hasAskedToAdjustSettings = true;
                    return;
                }
                OnAddVolumeTriggeredMorphButtonClicked();
            });
            button.height = 70f;
        }

        private void RefreshVolumeTriggeredMorphsUI()
        {
            var triggerCollection = ActiveCollection.GetVolumeTrigger().VolumeTriggeredMorphActions();
            if (triggerCollection == null) return;
            ResetVolumeTriggeredMorphsUI();
            foreach (var triggerAction in triggerCollection)
            {
                CreateVolumeTriggeredMorphActionUI(triggerAction);
            }

            var morphSelection = C.MorphChooserJSON.val;
            if (morphSelection != "None" && triggerCollection.FindIndex(x => x.receiverTargetName == morphSelection) == -1)
            {
                CreateAddVolumeTriggeredMorphActionButton();
            }
        }

        private void ResetVolumeTriggeredMorphsUI()
        {
            foreach (var uiElem in _volumeTriggerUI)
            {
                uiElem.Destroy();
            }
            _volumeTriggerUI.Clear();
            if ((Object) _addVolumeTriggeredMorphActionButton == (Object) null) return;
            _addVolumeTriggeredMorphActionButton.Destroy();
            _addVolumeTriggeredMorphActionButton = null;
        }
        #endregion

        #region Callbacks
        protected override void OnActiveCollectionSelected(CollectionManager.ActiveCollectionSelectedEventArgs args)
        {
            base.OnActiveCollectionSelected(args);
            var activeCollection = args.After;
            InitSettings(activeCollection);
            _needsRefresh = true;
        }

        public void OnAddVolumeTriggeredMorphButtonClicked()
        {
            if ((Object) Collections == (Object) null) return;
            if (ActiveCollection == null) return;
            var morphLabel = C.MorphChooserJSON.val;
            if (string.IsNullOrEmpty(morphLabel)) return;
            var actionName = Tools.GetVolumeTriggeredMorphLabel(morphLabel);

            ActiveCollection.GetVolumeTrigger().CreateVolumeTriggeredMorphAction(actionName, morphLabel);
            _needsRefresh = true;
        }
        #endregion

        private void FixedUpdate()
        {
            if (!_needsRefresh) return;
            _needsRefresh = false;
            RefreshCollisionTriggerActionsUI();
            RefreshVolumeTriggeredMorphsUI();
        }
    }
}
