﻿using System;
using System.Collections.Generic;
using System.Linq;
using MVR.FileManagementSecure;
using SimpleJSON;

namespace AudioMate
{
    public class ScreenSaveLoad : Screen
    {
        public new static readonly ScreenDefinition Definition = new ScreenDefinition(
            "SaveLoad",
            "Save / Load",
            typeof(ScreenSaveLoad));

        private const string _collectionSaveExt = "amc";
        private const string _sceneSaveExt = "json";

        private const string _collectionSaveFolder = "Custom\\Sounds\\AudioMate";
        private const string _sceneSaveFolder = "Saves\\scene";
        private const string _varTemplatePath = "Custom\\PluginData\\AudioMate\\AudioMate.CollectionPackage.1.dat";


        private const string _allCollections = "All Collections";

        private JSONStorableStringChooser _saveCollectionsJSON;
        private JSONStorableBool _includeTriggerActionsJSON;
        private JSONStorableBool _includeAudioClipsJSON;
        private JSONStorableBool _createMissingAudioSourcesJSON;


        private bool _merge = false;
        public override void _Init()
        {
            InitSaveUI();
            UI.CreateSpacer();
            InitLoadUI();
        }

        private void InitLoadUI()
        {
            UI.CreateHeader("Load Collection(s)", 2);
            _createMissingAudioSourcesJSON = new JSONStorableBool("Create Missing Audio Sources", true);
            UI.CreateToggle(_createMissingAudioSourcesJSON);
            var loadUI = UI.CreateButton("Load");
            loadUI.height = 60f;
            loadUI.button.onClick.AddListener(() => LoadAudioMateCollectionsFileDialog());

            var mergeLoadUI = UI.CreateButton("Load and Merge");
            mergeLoadUI.height = 60f;
            mergeLoadUI.button.onClick.AddListener(() => LoadAudioMateCollectionsFileDialog(true));


            if (C.DisplayHints) {
                UI.CreateText("Load a AudioMate setup from a file.", FontSizes.Small);
                UI.CreateText(
                    "If 'Create Audio Sources' is checked, all missing AudioSources that are not present in the current scene will be created.",
                    FontSizes.Small);
            }

            var sceneLoadUI = UI.CreateButton("Load from Scene File");
            sceneLoadUI.height = 60f;
            sceneLoadUI.button.onClick.AddListener(LoadAudioMateCollectionsFromScene);
        }

        private void InitSaveUI()
        {
            UI.CreateHeader("Save Collection(s)", 2);

            _saveCollectionsJSON = new JSONStorableStringChooser(
                "Collection",
                new List<string> { _allCollections }.Concat(C.collections.Names)
                    .ToList(),
                _allCollections,
                "Collection")
            {
                isStorable = false
            };
            var exportCollectionsUI = UI.CreatePopup(_saveCollectionsJSON, true);
            exportCollectionsUI.popupPanelHeight = 800f;

            _includeTriggerActionsJSON = new JSONStorableBool("Include Trigger Actions", true);
            UI.CreateToggle(_includeTriggerActionsJSON);
            _includeAudioClipsJSON = new JSONStorableBool("Include Audio Files", true);
            UI.CreateToggle(_includeAudioClipsJSON);

            var exportUI = UI.CreateButton("Save", Styles.Success);
            exportUI.height = 60f;
            exportUI.button.onClick.AddListener(() => SaveAudioMateCollectionsFileDialog());
            if (!C.DisplayHints) return;
            UI.CreateText("Save this Atom's AudioMate setup to a file. Those files can then be loaded by other AudioMate instances.", FontSizes.Small);
            UI.CreateText("The option 'Include Trigger Actions' adds all TriggerActions triggering your Collections to the saved setup.", FontSizes.Small);
            UI.CreateText("'Include Audio Files' will result in a new folder containing a copy of all needed audio files.", FontSizes.Small);
        }

        private void SaveAudioMateCollectionsFileDialog(bool legacy = false)
        {
            try
            {
                // Check if save directory exists. Create it if it doesn't.
                if (!FileManagerSecure.DirectoryExists(_collectionSaveFolder)) FileManagerSecure.CreateDirectory(_collectionSaveFolder);
                var fileBrowserUI = SuperController.singleton.fileBrowserUI;
                fileBrowserUI.SetTitle("Save collection(s)");
                fileBrowserUI.fileRemovePrefix = null;
                fileBrowserUI.hideExtension = false;
                fileBrowserUI.keepOpen = false;
                fileBrowserUI.fileFormat = _collectionSaveExt;
                fileBrowserUI.defaultPath = _collectionSaveFolder;
                fileBrowserUI.showDirs = true;
                fileBrowserUI.shortCuts = null;
                fileBrowserUI.browseVarFilesAsDirectories = false;
                fileBrowserUI.SetTextEntry(true);
                fileBrowserUI.Show(SaveAudioMateCollectionsFile);
                fileBrowserUI.ActivateFileNameField();
            }
            catch (Exception e)
            {
                SuperController.LogError($"AudioMate: Initialising save dialogue failed: {e}");
            }
        }

        private void SaveAudioMateCollectionsFile(string path)
        {
            if (string.IsNullOrEmpty(path)) return;
            var amcPath = !path.ToLower().EndsWith($".{_collectionSaveExt}") ? $"{path}.{_collectionSaveExt}" : path;
            var clipsPath = GetClipsPath(amcPath);
            var all = _saveCollectionsJSON.val == _allCollections;

            try
            {
                var json = new JSONClass {{Storables.Collections, null}};
                var collectionName = all ? null : _saveCollectionsJSON.val;

                if (_includeAudioClipsJSON.val)
                {
                    if (!FileManagerSecure.DirectoryExists(clipsPath)) FileManagerSecure.CreateDirectory(clipsPath);
                    var clips = C.collections.UsedClips(all ? null : _saveCollectionsJSON.val);
                    foreach (var clip in clips.Where(clip => clip.SourcePath != null && clip.AudioClip != null))
                    {
                        var newPath = FileManagerSecure.NormalizePath($"{clipsPath}\\{clip.AudioClip.displayName}");
                        var sourcePath = FileManagerSecure.NormalizePath(clip.SourcePath);
                        if (sourcePath == newPath) continue;
                        if (FileManagerSecure.FileExists(newPath)) FileManagerSecure.DeleteFile(newPath);
                        FileManagerSecure.CopyFile(sourcePath, newPath);
                        clip.SourcePath = null;
                    }
                }
                C.collections.AddToJSON(json, Storables.Collections, collectionName, !_includeTriggerActionsJSON.val);
                json[Storables.IncludesAudioClips] = _includeAudioClipsJSON.val ? "true" : "false";
                SuperController.singleton.SaveJSON(json, amcPath);
                // TODO: Screenshots
                // SuperController.singleton.DoSaveScreenshot(path.Replace($".{_collectionSaveExt}", ".jpg"));
                // TODO: package .amc file and clips into one file
            }
            catch (Exception e)
            {
                SuperController.LogError($"AudioMate: Exporting collection failed: {e}");
            }
        }

        private void LoadAudioMateCollectionsFileDialog(bool merge = false)
        {
            _merge = merge;
            try
            {
                if (!FileManagerSecure.DirectoryExists(_collectionSaveFolder)) FileManagerSecure.CreateDirectory(_collectionSaveFolder);
                var shortcuts = FileManagerSecure.GetShortCutsForDirectory(_collectionSaveFolder);
                SuperController.singleton.GetMediaPathDialog(LoadAudioMateCollectionsFile, _collectionSaveExt, _collectionSaveFolder, false, true, false, null, false, shortcuts);
            }
            catch (Exception exc)
            {
                SuperController.LogError($"AudioMate: Initialising load dialogue failed: {exc}");
            }
        }

        private void LoadAudioMateCollectionsFile(string path)
        {
            if (string.IsNullOrEmpty(path)) return;
            try
            {
                var json = SuperController.singleton.LoadJSON(path);

                if (!json.Childs.Any())
                {
                    SuperController.LogError($"AudioMate.ScreenSaveLoad.LoadAudioMateCollectionsFile: {path} content is invalid.");
                    return;
                }

                if (json[Storables.Collections] == null || !json[Storables.Collections].Childs.Any())
                {
                    SuperController.LogError($"AudioMate.ScreenSaveLoad.LoadAudioMateCollectionsFile: Loaded JSON has no collection data.");
                    return;
                }

                if (json[Storables.IncludesAudioClips] != null && json[Storables.IncludesAudioClips].Value == "true")
                {
                    var clipsPath = GetClipsPath(path);
                    if (!FileManagerSecure.DirectoryExists(clipsPath))
                    {

                        return;
                    }
                    C.audioFileManager.ImportFolder(clipsPath, () =>
                    {
                        C.collections.RestoreFromJSON(json[Storables.Collections], _merge, _createMissingAudioSourcesJSON.val);
                    });
                }
                else
                {
                    C.collections.RestoreFromJSON(json[Storables.Collections], _merge, _createMissingAudioSourcesJSON.val);
                }

            }
            catch (Exception e)
            {
                SuperController.LogError($"AudioMate.{nameof(ScreenSaveLoad)}.{nameof(LoadAudioMateCollectionsFile)}: Failed to load collection {path}: {e}");
            }
        }

        private string GetClipsPath(string path)
        {
            return $"{path}.clips";
        }

        private void LoadAudioMateCollectionsFromScene()
        {
            try
            {
                FileManagerSecure.CreateDirectory(_sceneSaveFolder);
                var shortcuts = FileManagerSecure.GetShortCutsForDirectory(_sceneSaveFolder);
                SuperController.singleton.GetMediaPathDialog(LoadAudioMateCollectionsFromSceneDialog, _sceneSaveExt, _sceneSaveFolder, false, true, false, null, false, shortcuts);
            }
            catch (Exception exc)
            {
                SuperController.LogError($"AudioMate: Initialising load scene dialogue failed: {exc}");
            }
        }

        private void LoadAudioMateCollectionsFromSceneDialog(string path)
        {

            if (string.IsNullOrEmpty(path)) return;

            try
            {
                var candidates = new List<JSONClass>();
                var json = SuperController.singleton.LoadJSON(path);
                if (!json.Childs.Any())
                {
                    SuperController.LogError($"AudioMate: content of {path} is invalid.");
                    return;
                }

                if (json["atoms"] != null)
                {
                    for (var i = 0; i < json["atoms"].AsArray.Count; i++)
                    {
                        var entry = json["atoms"][i];
                        if (entry["storables"] == null) continue;
                        foreach (JSONNode node in entry["storables"].AsArray)
                        {
                            if (!(node["id"] != null & (node["id"].Value.Contains("_AudioMate.AudioMateController") || node["id"].Value == Variables.Plugin)))
                                continue;

                            if (node[Storables.Collections] != null)
                                candidates.Add(node[Storables.Collections].AsObject);
                        }
                    }
                }

                var found = candidates.Count;
                switch (found)
                {
                    case 0:
                        UI.CreateModal("Found no AudioMate instance in the loaded scene. Please make sure you picked the right file.", OverlayTemplates.Ok);
                        break;
                    case 1:
                        C.collections.RestoreFromJSON(candidates[0], _merge, _createMissingAudioSourcesJSON.val);
                        break;
                    default:
                    {
                        if (found > 1)
                        {
                            var callbacks = new Action[] { () => C.collections.RestoreFromJSON(candidates[0], _merge, _createMissingAudioSourcesJSON.val) };
                            UI.CreateModal(
                                "Found {found} AudioMate instances. Importing more than one instance is not implemented yet. Do you want to import the first instance?",
                                OverlayTemplates.YesNo,
                                callbacks);
                        }

                        break;
                    }
                }
            }
            catch (Exception e)
            {
                SuperController.LogError($"AudioMate.{nameof(ScreenSaveLoad)}.{nameof(LoadAudioMateCollectionsFile)}: Failed to load collection {path}: {e}");
            }
        }

    }
}
