﻿using System;
using System.Collections.Generic;
using System.Linq;
using UnityEngine;
using UnityEngine.UI;
using Object = UnityEngine.Object;

namespace AudioMate
{

    public class UIActiveCollection : UIComponent
    {
        private static UIFactory _ui;

        public static UIActiveCollection AddTo(Transform parent)
        {
            var go = new GameObject();
            go.transform.SetParent(parent, false);

            var layout = go.AddComponent<LayoutElement>();
            layout.flexibleHeight = 1;
            layout.flexibleWidth = 0;
            layout.preferredWidth = 425f;
            return Configure(go);
        }

        private static UIActiveCollection Configure(GameObject go)
        {
            var group = go.AddComponent<VerticalLayoutGroup>();
            group.spacing = 10f;
            group.childForceExpandHeight = false;
            group.childControlHeight = true;
            group.childForceExpandWidth = true;
            group.childControlWidth = true;
            var rect = go.GetComponent<RectTransform>() ?? go.AddComponent<RectTransform>();
            rect.pivot = new Vector2(0, 1);
            rect.StretchParent();
            _ui = go.AddComponent<UIFactory>();
            return go.AddComponent<UIActiveCollection>();
        }

        private UIDynamicPopup _popupCollectionList;

        private JSONStorableString _collectionNameJSON;
        private JSONStorableString _collectionContentJSON;


        private UIDynamicTextField _activeCollectionClipListText;

        private Collection _activeCollection;

        private JSONStorableStringChooser _receivingAtomJSON;
        private JSONStorableStringChooser _receivingNodeJSON;

        private UIDynamicButton _btnClear;
        private UIDynamicButton _btnPlayClip;
        private UIDynamicButton _btnQueueClip;
        private UIDynamicButton _btnStop;

        protected override void _Init()
        {
            _receivingAtomJSON = _c.ReceivingAtomJSON;
            _receivingNodeJSON = _c.ReceivingNodeJSON;
            _ui.c = _c;
            CreateCollectionChooserUI();
            CreateCollectionManagementUI();
            _ui.CreateSpacer().height = 15f;
            CreateActiveCollectionNameUI();
            CreateAudioReceiverUI();
            CreateActiveCollectionContentUI();
            CreateActiveCollectionPreviewUI();
            isInitialized = true;
        }

        protected override void _Bind()
        {
            if (_c.collections.ActiveCollection == null)
            {
                Log("WARNING: Can not bind Active Collection UI Events because Active Collection is null.");
                return;
            }

            Log("Binding Active Collection UI Events");
            BindActiveCollection(_c.collections.ActiveCollection);
            _c.clipLibrary.OnClipsRefreshed.AddListener(OnClipLibraryUpdated);
            _popupCollectionList.popup.onOpenPopupHandlers += _c.collections.SyncCollectionNames;
            _btnClear.button.onClick.AddListener(_c.collections.ClearActiveCollection);
            _btnPlayClip.button.onClick.AddListener(() =>
            {
                _c.collections.PlayClipAction();
            });
            _btnQueueClip.button.onClick.AddListener(() =>
            {
                _c.collections.QueueClipAction();
            });
            _btnStop.button.onClick.AddListener(() =>
            {
                var audioSource = _c.collections.ActiveCollection.ReceiverAudioSource;
                if ((Object) audioSource == (Object) null) return;
                audioSource.Stop();
                var lastPlayedClip = _c.Cursor.GetClip();
                var testAudioSource = lastPlayedClip?.AudioClip?.manager.testAudioSource;
                if ((Object) testAudioSource != (Object) null)
                {
                    testAudioSource.Stop();
                }
                _c.collections.ActiveCollection.HasPlayedAnything = false;
            });
            _c.collections.SyncCollectionNames();
            _c.collections.SelectActiveCollection();
        }

        private void CreateCollectionChooserUI()
        {
            _popupCollectionList = _ui.CreatePopup(_c.CollectionsJSON, true);
        }

        private void CreateCollectionManagementUI()
        {
            var btnGroup = _ui.CreateHorizontalGroup();
            btnGroup.CreateButton(Symbols.Add, Styles.Button,
                () => _c.collections.AddCollection());
            _btnClear = btnGroup.CreateButton("Clear");
            btnGroup.CreateButton(Symbols.Remove, Styles.Danger,
                () => _c.collections.RemoveActiveCollection());
        }

        private void CreateActiveCollectionNameUI()
        {
            _collectionNameJSON = new JSONStorableString("Collection Name", null, UpdateCollectionName);
            _ui.CreateSimpleTextInput(_collectionNameJSON, 70);
        }

        private void CreateActiveCollectionContentUI()
        {
            _collectionContentJSON = new JSONStorableString("Collection Content", "", UpdateCollectionContent);
            _activeCollectionClipListText = _ui.CreateTextField(_collectionContentJSON);
            _activeCollectionClipListText.height = 425f;
        }

        private void CreateActiveCollectionPreviewUI()
        {
            var btnGroupPreview = _ui.CreateHorizontalGroup();
            _btnPlayClip = btnGroupPreview.CreateButton($"{Symbols.Play}", Styles.Success, null, true);
            _btnQueueClip = btnGroupPreview.CreateButton("Queue");
            _btnStop = btnGroupPreview.CreateButton($"{Symbols.Stop}");
        }

        private void UpdateCollectionName(string val)
        {
            if (string.IsNullOrEmpty(val))
            {
                _collectionNameJSON.valNoCallback = _activeCollection.Name;
                return;
            }

            var finalName = _c.collections.GetAvailableCollectionName(val);
            if (finalName != val)
            {
                _collectionNameJSON.valNoCallback = finalName;
            }
            _c.collections.SetActiveCollectionName(finalName);
        }

        private void UpdateCollectionContent(string val)
        {
            _activeCollectionClipListText.text = _activeCollection.ToString();
        }

        private void CreateAudioReceiverUI()
        {
            if (_receivingAtomJSON == null || _receivingNodeJSON == null) return;

            var receiverAtomUI = _ui.CreatePopup(_receivingAtomJSON, true);
            receiverAtomUI.popup.onOpenPopupHandlers += SyncReceiverAtomChoices;
            SyncReceiverAtomChoices();

            var receiverNodeUI = _ui.CreatePopup(_receivingNodeJSON, true);
            receiverNodeUI.popup.onOpenPopupHandlers += SyncReceiverNodeChoices;
            SyncReceiverNodeChoices();
        }
        private void SyncReceiverAtomChoices()
        {
            var choices = new List<string> {"None"};

            choices.AddRange(from atom in SuperController.singleton.GetAtoms()
                where atom != null
                where atom.category == "People" || atom.category == "Sound"
                select atom.name);

            _receivingAtomJSON.choices = choices;
        }

        private void SyncReceiverNodeChoices()
        {
            var choices = new List<string> {"None"};

            if (_receivingAtomJSON == null || string.IsNullOrEmpty(_receivingAtomJSON?.val) ||
                _receivingAtomJSON?.val == "None") return;

            // Get nodes
            var receivers = SuperController.singleton.GetAtomByUid(_receivingAtomJSON.val)?.GetStorableIDs();

            if (receivers == null) return;

            choices.AddRange(receivers);

            _receivingNodeJSON.choices = choices;
        }

        private void BindActiveCollection(Collection collection)
        {
            if (collection == null) return;
            if (_activeCollection != null) {
                if (collection.Name == _activeCollection.Name) return;
            }
            _activeCollection = collection;
            _activeCollection.OnClipPlayed.AddListener(RefreshActiveCollectionClipList);
            if (_popupCollectionList.popup.currentValue != _activeCollection.Name)
            {
                _popupCollectionList.popup.currentValue = _activeCollection.Name;
            }

            if (_collectionNameJSON.val != _activeCollection.Name) _collectionNameJSON.valNoCallback = _activeCollection.Name;
            RefreshActiveCollectionClipList();
        }

        public override void OnActiveCollectionSelected(CollectionManager.ActiveCollectionSelectedEventArgs eventArgs)
        {
            BindActiveCollection(eventArgs.After);
            _collectionNameJSON.valNoCallback = eventArgs.After.Name;

            OnActiveCollectionUpdated();
        }

        public override void OnActiveCollectionUpdated()
        {
            RefreshActiveCollectionClipList();
        }

        public void OnClipLibraryUpdated()
        {
            RefreshActiveCollectionClipList();
        }

        public override void OnContainingAtomNameChanged(AudioMateController.ContainingAtomNameChangedEventArgs eventArgs)
        {
            if (_receivingAtomJSON.val == eventArgs.Before) _receivingAtomJSON.val = eventArgs.After;
            SyncReceiverAtomChoices();
        }

        public void RefreshActiveCollectionClipList()
        {
            if (_activeCollection == null)
            {
                Log("Active Collection is null. Aborting.");
                return;
            }
            if ((Object) _activeCollectionClipListText == (Object) null)
            {
                Log("Active Collection Clip List Text is null. Aborting.");
                return;
            }
            try
            {
                var lastPlayed = _activeCollection.CurrentClip ?? _activeCollection.FirstOrDefault();
                var text = "";
                if (_activeCollection.Count > 0)
                {
                    _activeCollectionClipListText.backgroundImage.color = Styles.EmbeddedBg;
                    _activeCollectionClipListText.textColor = Styles.EmbeddedText;

                    foreach (AudioMateClip clip in _activeCollection)
                    {
                        var clipName = $"{(clip.IsProvisional ? "Loading: " : "")}{clip.DisplayName}";
                        text += lastPlayed != null & clip == lastPlayed
                            ? $"<b>{clipName}</b>"
                            : clipName;
                        text += "\n";
                    }
                }
                else
                {
                    _activeCollectionClipListText.backgroundImage.color = Color.clear;
                    _activeCollectionClipListText.textColor = Color.white;
                    _activeCollectionClipListText.UItext.alignment = TextAnchor.MiddleCenter;
                    text = "\n"
                           + "\n"
                           + "<b>Empty</b>".ToUpper()
                           + "\n"
                           + "\n"
                           + "Add clips to this <b>Collection</b> by selecting audio clips in the <b>Clip Library</b>.";
                }

                _activeCollectionClipListText.text = text;

            }
            catch (Exception e)
            {
                Log($"RefreshActiveCollectionClipList: {e}");
            }
        }

        protected override void Log(string message)
        {
            if ((Object) _c == (Object) null) return;
            _c.Log($" [UIActiveCollection]: {message}");
        }
    }
}
