﻿using System;
using System.Collections.Generic;
using System.Linq;
using UnityEngine;
using UnityEngine.Serialization;
using UnityEngine.UI;
using Object = UnityEngine.Object;

namespace AudioMate
{

    public class UIClipLibrary : UIComponent
    {
        private static ScrollRect _content;

        public static UIClipLibrary AddTo(Transform parent)
        {
            var go = new GameObject();
            go.transform.SetParent(parent, false);
            var rect = go.AddComponent<RectTransform>();
            rect.StretchParent();
            var layout = go.AddComponent<LayoutElement>();
            layout.preferredHeight = 825f;
            _content = UIFactory.CreateScrollRect(go, new RectOffset(0, 0, 0, 0));
            var rectTransform = _content.content.GetComponent<RectTransform>();
            rectTransform.StretchParent();
            return Configure(go);
        }

        private static UIClipLibrary Configure(GameObject go)
        {
            var container = new GameObject();
            container.transform.SetParent(go.transform, false);
            var group = container.AddComponent<VerticalLayoutGroup>();
            group.spacing = 10f;
            group.childForceExpandHeight = true;
            group.childControlHeight = false;
            group.childForceExpandWidth = true;
            group.childControlWidth = false;
            var rect = container.GetComponent<RectTransform>();
            rect.pivot = new Vector2(0, 1);
            rect.StretchParent();

            return container.AddComponent<UIClipLibrary>();
        }

        [FormerlySerializedAs("isRefreshing")] public bool needsRefresh;

        public string activeFilter;

        private bool _isRefreshing;

        protected override void _Init()
        {
            _c.clipLibrary.OnClipsRefreshed.AddListener(OnClipsRefreshed);
            _content.onValueChanged.AddListener(OnScroll);
            Refresh();
        }

        public void OnClipsRefreshed()
        {
            Refresh();
        }

        public void Refresh()
        {
            needsRefresh = true;
        }

        private void RefreshUI()
        {
            if (_isRefreshing) return;
            Log("Refreshing Clip Library UI");
            if (_c.Clips == null || _c.Clips.Count <= 0)
            {
                Log("No Clips found. Aborting BuildClipUI.");
                return;
            }

            try
            {
                _isRefreshing = true;
                var activeCollection = _c.collections.ActiveCollection;
                var clipsToDestroy = new List<AudioMateClip>();
                foreach (var clip in _c.Clips)
                {
                    // Destroy clips without valid scene audio file
                    if ((!clip.IsProvisional && clip.AudioClip == null) || clip.AudioClip?.destroyed == true)
                    {
                        clipsToDestroy.Add(clip);
                        continue;
                    }

                    // Filter by collection if filter is active
                    if (activeFilter != ScreenClips.FilterAll)
                    {
                        if ((activeFilter == ScreenClips.FilterAssigned && !activeCollection.Contains(clip))
                            || (activeFilter == ScreenClips.FilterUnassigned && activeCollection.Contains(clip)))
                        {
                            clip.Hide();
                            continue;
                        }
                    }

                    // Only create a new Clip UI if it doesn't exist, isn't initialised or hasn't been added to the
                    // tab view.
                    if (clip.UI == null || clip.UI.IsInitialised == false)
                    {
                        clip.CreateClipUI(
                            _content.content,
                            () =>
                            {
                                clip.AudioClip?.Test();
                                _c.Cursor.SetClip(clip);
                            },
                            () =>
                            {
                                if (!_c.isBulkSelectOn)
                                {
                                    clip.ToggleState();
                                    if (clip.IsInActiveCollection)
                                    {
                                        _c.collections.AddClipToActiveCollection(clip);
                                    }
                                    else
                                    {
                                        _c.collections.RemoveClipFromActiveCollection(clip);
                                    }

                                    _c.Cursor.SetClip(clip);
                                }
                                else
                                {
                                    _c.collections.ToggleClipsInActiveCollection(clip.CurrentIndex,
                                        _c.bulkSelectAmount);
                                    _c.isBulkSelectOn = false;
                                    _c.Cursor.SetClip(clip);
                                }
                            },
                            () => OnDeleteClipFromScene(clip));
                        clip.OnMouseEntered.AddListener(OnMouseEnteredClipUI);
                        clip.OnMouseExited.AddListener(OnMouseExitedClipUI);
                    }

                    clip.Show();
                    // Set the clip state. It shows if a clip is assigned to the active collection.
                    clip.SetState(activeCollection.Contains(clip));
                    clip.ShowRemoveButton(_c.AdvancedUIEnabledJSON.val);
                }

                // Destroy Clips and remove them from Clip Library
                foreach (var clip in clipsToDestroy)
                {
                    if (clip != null) _c.Clips.Remove(clip);
                    clip?.Destroy();
                }

                for (var i = 0; i < _c.Clips.Count; i++)
                {
                    if (_c.Clips[i] != null) _c.Clips[i].CurrentIndex = i;
                }
                if (_c.AddImportedFilesToActiveCollectionJSON.val)
                {
                    if (_c.audioFileManager.AddTheseFilesToActiveCollection.Count <= 0) return;
                    for (var index = 0; index < _c.audioFileManager.AddTheseFilesToActiveCollection.Count; index++)
                    {
                        var namedAudioClip = _c.audioFileManager.AddTheseFilesToActiveCollection[index];
                        foreach (var audioMateClip in _c.Clips)
                        {
                            if (audioMateClip.AudioClip != namedAudioClip) continue;
                            _c.collections.AddClipToActiveCollection(audioMateClip);
                            _c.audioFileManager.AddTheseFilesToActiveCollection.Remove(namedAudioClip);
                        }
                    }
                }
                else
                {
                    if (_c.audioFileManager.AddTheseFilesToActiveCollection.Count > 0) _c.audioFileManager.AddTheseFilesToActiveCollection.Clear();
                }
            }
            catch (Exception e)
            {
                Log($"{e})");
            }
            finally
            {
                _isRefreshing = false;
            }
        }

        private void OnMouseEnteredClipUI(AudioMateClip.ClipMouseOverStateChangedEventArgs args)
        {
            if (!_c.isBulkSelectOn || _c.bulkSelectAmount <= 0)
            {
                return;
            }
            var clip = args.Clip;
            var fromIndex = _c.Clips.IndexOf(clip);
            var toIndex = fromIndex + _c.bulkSelectAmount - 1;
            var index = 0;
            foreach (var audioMateClip in _c.Clips)
            {
                if (index >= fromIndex && index <= toIndex)
                {
                    audioMateClip.PreviewToggleState();
                }
                index++;
            }
        }

        private void OnMouseExitedClipUI(AudioMateClip.ClipMouseOverStateChangedEventArgs args)
        {
            if (!_c.isBulkSelectOn || _c.bulkSelectAmount <= 0)
            {
                return;
            }
            foreach (var audioMateClip in _c.Clips)
            {
                audioMateClip.ResetToggleStatePreview();
            }
        }

        private Vector2 _lastScrollValue;
        private void OnScroll(Vector2 value)
        {
            if (!(Math.Abs(value.x - _lastScrollValue.x) > 0.1f)) return;
            _lastScrollValue = value;
            foreach (var audioMateClip in _c.Clips)
            {
                audioMateClip.UI.MouseOver.Refresh();
            }
        }

        public void SetFilter(string filterMode)
        {
            activeFilter = filterMode;
            Refresh();
        }

        private void OnDeleteClipFromScene(AudioMateClip clip)
        {
            if (_c.collections.UsedClips().Contains(clip))
            {
                _c.ui.ShowYesNoModal("This clip is assigned to a collection. Are you sure that you want to delete it from the scene?",
                    () =>
                    {
                        _c.collections.RemoveClipFromAllCollections(clip);
                        clip.DeleteFromScene();
                    }, null, "Warning");
                return;
            }
            clip.DeleteFromScene();
            Refresh();
        }

        public override void OnActiveCollectionSelected(CollectionManager.ActiveCollectionSelectedEventArgs eventArgs)
        {
            Refresh();
        }

        public override void OnActiveCollectionUpdated()
        {
            Refresh();
        }

        protected override void Log(string message)
        {
            if ((Object) _c == (Object) null) return;
            _c.Log($" [UIClipLibrary]: {message}");
        }

        protected void FixedUpdate()
        {
            if (!needsRefresh || _isRefreshing) return;
            needsRefresh = false;
            RefreshUI();
        }

        public override void OnDestroy()
        {
            foreach (var audioMateClip in _c.Clips)
            {
                audioMateClip.DestroyUI();
            }

            base.OnDestroy();
        }
    }
}
