﻿using System;
using UnityEngine;
using UnityEngine.Events;
using UnityEngine.Serialization;
using UnityEngine.UI;
using Object = UnityEngine.Object;

namespace AudioMate
{
    [RequireComponent(typeof(UIDynamicButton))]
    public class UIManager : MonoBehaviour
    {
        public static UIManager AddTo(Transform transform)
        {
            var go = new GameObject();
            go.transform.SetParent(transform, false);

            var topLevelFactory = go.AddComponent<UIFactory>(); ;


            var tabs = Tabs.AddTo(go.transform);

            var rect = go.AddComponent<RectTransform>();
            rect.StretchTop();
            rect.pivot = new Vector2(0, 1);

            go.AddComponent<VerticalLayoutGroup>();

            var panels = new GameObject();
            panels.transform.SetParent(rect.transform, false);

            var panelsGroup = panels.AddComponent<HorizontalLayoutGroup>();
            panelsGroup.spacing = 10f;
            panelsGroup.childControlWidth = true;
            panelsGroup.childForceExpandWidth = false;
            panelsGroup.childControlHeight = true;
            panelsGroup.childForceExpandHeight = true;

            var leftPanel = CreatePanel(panels.transform, 0f, 1f);
            var rightPanel = CreatePanel(panels.transform, 450f, 0f, true);

            var ui = go.AddComponent<UIManager>();
            ui.mainPanel = topLevelFactory;
            ui.tabs = tabs;
            ui.leftPanel = leftPanel;
            ui.rightPanel = rightPanel;
            return ui;
        }

        private static GameObject CreatePanel(Transform transform, float preferredWidth, float flexibleWidth, bool embedded = false)
        {
            var go = new GameObject();
            go.transform.SetParent(transform, false);

            var rect = go.AddComponent<RectTransform>();
            rect.pivot = new Vector2(0, 1f);
            var layout = go.AddComponent<LayoutElement>();
            layout.minWidth = 0;
            layout.preferredWidth = preferredWidth;
            layout.flexibleWidth = flexibleWidth;
            layout.flexibleHeight = 1;

            if (embedded)
            {
                var image = go.gameObject.AddComponent<Image>();
                image.color = Styles.EmbeddedBg;
                image.preserveAspect = false;
                var verticalLayoutGroup = go.AddComponent<VerticalLayoutGroup>();
                verticalLayoutGroup.childControlWidth = true;
                verticalLayoutGroup.childForceExpandWidth = true;
                verticalLayoutGroup.childControlHeight = true;
                verticalLayoutGroup.childForceExpandHeight = false;
                verticalLayoutGroup.spacing = 10f;
                verticalLayoutGroup.padding = new RectOffset(20, 20, 20, 20);
            }
            else
            {
                var verticalLayoutGroup = go.AddComponent<VerticalLayoutGroup>();
                verticalLayoutGroup.childControlWidth = true;
                verticalLayoutGroup.childForceExpandWidth = true;
                verticalLayoutGroup.childControlHeight = true;
                verticalLayoutGroup.childForceExpandHeight = false;
                verticalLayoutGroup.spacing = 5f;
            }



            return go;
        }

        public class UIInitializedEvent : UnityEvent{};
        public readonly UIInitializedEvent OnUIInitialized = new UIInitializedEvent();

        public ScreenManager screenManager;
        public bool isInitialized;
        private AudioMateController _c;

        [FormerlySerializedAs("container")] public UIFactory mainPanel;
        public GameObject leftPanel;
        public GameObject rightPanel;
        public Tabs tabs;
        public UIActiveCollection activeCollectionUI;


        #region Init
        public void Init(AudioMateController mainController)
        {
            if ((Object) mainController == (Object) null)
            {
                SuperController.LogError("AudioMate: Error during UI initialization. Controller or Collection Manager are set to null.");
                return;
            }
            Log("Initializing UI Manager");
            _c = mainController;
            mainPanel.c = _c;
            screenManager = InitScreenManager(leftPanel);
            InitActiveCollectionUI();
        }

        private static ScreenManager InitScreenManager(GameObject panel)
        {
            var go = new GameObject();
            go.transform.SetParent(panel.transform, false);

            var layout =  go.AddComponent<LayoutElement>();
            layout.flexibleHeight = 1;
            layout.flexibleWidth = 1;

            var screensManager = ScreenManager.AddTo(go);
            return screensManager;
        }

        public void InitActiveCollectionUI()
        {
            if ((Object) _c == (Object) null) return;
            activeCollectionUI = UIActiveCollection.AddTo(rightPanel.transform);
            activeCollectionUI.Init(_c);
        }

        public void Bind()
        {
            if ((Object) _c == (Object) null)
            {
                SuperController.LogError("AudioMate: Error during bind. Can't access main controller.");
                return;
            }
            try
            {
                Log("Binding Main UI Events");

                // Screens are defined in ScreenManager.cs
                Screens.List.ForEach(screen => tabs.Add(screen.ID, screen.Title));

                screenManager.OnScreenChanged.AddListener(screenID => tabs.Select(screenID));
                screenManager.Bind(_c);

                tabs.OnTabSelected.AddListener(screenManager.ChangeScreen);

                activeCollectionUI.Bind();
            }
            catch (Exception e)
            {
                SuperController.LogError($"AudioMate.{nameof(UIManager)}.{nameof(Bind)}: {e}");
            }
        }
        #endregion

        public void ShowOkModal(string message, Action callback = null, string heading = null)
        {
            var callbacks = callback != null ? new [] { callback } : null;
            mainPanel.CreateModal(message, OverlayTemplates.Ok, callbacks, heading);
        }

        public void ShowYesNoModal(string message, Action yesCallback = null, Action noCallback = null, string heading = null)
        {
            var callbacks = new [] { yesCallback, noCallback };
            mainPanel.CreateModal(message, OverlayTemplates.YesNo, callbacks, heading);
        }

        public void FixedUpdate()
        {
            if (isInitialized || !activeCollectionUI.isInitialized || !screenManager.isInitialized) return;
            isInitialized = true;
            OnUIInitialized.Invoke();
        }

        #region Callbacks
        public void OnDestroy()
        {
            if (rightPanel != null) Destroy(rightPanel);
            if (leftPanel != null) Destroy(leftPanel);
        }
        #endregion

        #region Helpers
        private void Log(string message)
        {
            if ((Object) _c == (Object) null) return;
            _c.Log($" [UIManager]: {message}");
        }
        #endregion
    }
}
